#include "frmBitmap.h"

#include <Application.h>
#include <OS.h>
#include <Region.h>
#include <GraphicsDefs.h>
#include <Screen.h>

#include <new>

#include <memory.h>

#include "frmHelper.h"

#define	MSG_FLIP			'FLIP'

// view ///////////////////////////////////////////////////////////////////////////

FlipView::FlipView(FrameworkParams &params, status_t &error, const char * &error_text) :

	BView(BRect(B_ORIGIN, params.frameBounds), NULL, B_FOLLOW_NONE, B_WILL_DRAW),
	
	mRing(NULL)
{
	if (B_OK != error)
		return;

	mMinimized = true;
	mRingSize = params.numRingBuffers;
	mFlipSelector = mRingSize - 1;

	uint32 bitmapsSupport = 0; 

	if (! ::bitmaps_support_space(params.pixelFormat, &bitmapsSupport) ||
		! bitmapsSupport & B_VIEWS_SUPPORT_DRAW_BITMAP)
	{
		error_text = "unsupported bitmap color space.";
		error = B_ERROR;
		return;
	}

	mRing = new(std::nothrow) BufferRecord [mRingSize];
	if (NULL == mRing)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	for (int32 i = 0; i < mRingSize; i++)
	{
		mRing[i].writeLock = ::kreate_sem(0, NULL);
		mRing[i].unlockAtActivate = true;
		
		if (B_OK > mRing[i].writeLock)
		{
			error_text = "failure providing ring locking mechanism at framework construction.";
			error = mRing[i].writeLock;
			break;
		}
		
		mRing[i].bitmap = new(std::nothrow) BBitmap(BRect(B_ORIGIN, params.frameBounds), params.pixelFormat);
		if (NULL == mRing[i].bitmap)
		{
			error_text = "'bad_alloc' at framework construction.";
			error = B_ERROR;
			break;
		}
	}

	if (B_OK != error)
		return;

	SetViewColor(B_TRANSPARENT_COLOR);
}

FlipView::~FlipView()
{
	if (NULL != mRing)
	{
		for (int i = 0; i < mRingSize; i++)
		{
			::delete_sem(mRing[i].writeLock);
			delete mRing[i].bitmap;
		}

		delete [] mRing;
	}
}

void FlipView::DestroyRingLocks()
{
	if (NULL != mRing)
		for (int i = 0; i < mRingSize; i++)
		{
			::delete_sem(mRing[i].writeLock);
			mRing[i].writeLock = B_ERROR;
		}
}

void FlipView::MessageReceived(BMessage *message)
{
	switch (message->what)
	{
	case MSG_FLIP:

		++mFlipSelector %= mRingSize;
		// in this framework the flip selector is updated in advance and stays valid throughout
		// the lifespan of the frame so any draw()'s during that time can use that selector

		if (!mMinimized)
		{
			Invalidate();
			Window()->UpdateIfNeeded();

			::release_sem(mRing[mFlipSelector].writeLock);
		}
		else
			mRing[mFlipSelector].unlockAtActivate = true;

		break;

	default:
		BView::MessageReceived(message);
	} 
}

void FlipView::Draw(BRect updateRect)
{
	DrawBitmap(mRing[mFlipSelector].bitmap, updateRect, updateRect);
}

void FlipView::WindowActivated(bool)
{
	if (mMinimized != Window()->IsMinimized())
	{
		mMinimized = !mMinimized;

		if (!mMinimized)
			for (int i = 0; i < mRingSize; i++)
				if (mRing[i].unlockAtActivate)
				{
					mRing[i].unlockAtActivate = false;
					::release_sem_etc(mRing[i].writeLock, 1, B_DO_NOT_RESCHEDULE);
				}
	}
}

// window /////////////////////////////////////////////////////////////////////

FrmBmpWindow::FrmBmpWindow(FrameworkParams &params, status_t &error, const char * &error_text) :

	BWindow(BRect(B_ORIGIN, params.frameBounds), "BView_BBitmap Framework", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	mDrawSelector = 0;
	mView = NULL;

	if (B_OK != error)
		return;

	mView = new(std::nothrow) FlipView(params, error, error_text);
	if (NULL == mView)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	AddChild(mView);

	BScreen *screen = new(std::nothrow) BScreen(this);
	if (NULL == screen)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	BRect scrRect = screen->Frame();
	MoveTo((scrRect.Width() - params.frameBounds.x) * .5f, (scrRect.Height() - params.frameBounds.y) * .5f);
	delete screen;
}

FrmBmpWindow::~FrmBmpWindow()
{
	if (NULL != mView)
		mView->DestroyRingLocks();

	int32 result;
	thread_id painter = ::find_thread(Framework::__painter_name());

	if (B_NAME_NOT_FOUND != painter)
		::wait_for_thread(painter, &result);

	// note: all children of this window will be automatically deleted in BWindow's dtor.
}

bool FrmBmpWindow::QuitRequested()
{
	Hide(); 
	Sync();

 	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);

	return BWindow::QuitRequested();
}

status_t FrmBmpWindow::frame_open(Buffer &desc)
{
	status_t res = ::acquire_sem(mView->mRing[mDrawSelector].writeLock);

	if (B_OK == res)
	{
		desc.frameHeight	= mView->mRing[mDrawSelector].bitmap->Bounds().IntegerHeight() + 1;
		desc.frameWidth		= mView->mRing[mDrawSelector].bitmap->Bounds().IntegerWidth() + 1;
		desc.pixelFormat	= mView->mRing[mDrawSelector].bitmap->ColorSpace();
		desc.rowStride		= mView->mRing[mDrawSelector].bitmap->BytesPerRow();
		desc.bits			= mView->mRing[mDrawSelector].bitmap->Bits();
	}

	return res;
}

void FrmBmpWindow::frame_close()
{
	PostMessage(MSG_FLIP, mView);

	++mDrawSelector %= mView->mRingSize;
}
