#include "frmDirectWindowLC.h"

#include <Application.h>
#include <View.h>
#include <Screen.h>

#include <new>

#include <malloc.h>
#include <memory.h>

#include "memblt.h"
#include "frmHelper.h"

#define	MSG_FLIP			'FLIP'

// window ///////////////////////////////////////////////////////////////////////

FrmDirectWindowLC::FrmDirectWindowLC(FrameworkParams &params, status_t &error, const char * &error_text) :

	BDirectWindow(BRect(B_ORIGIN, params.frameBounds), "BDirectWindowLC Framework", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE),

	mRing(NULL),
	mBackBits(NULL)
{
	if (B_OK != error)
		return;

	mConnected = false;

	mRingSize = params.numRingBuffers;
	mDrawSelector = 0;
	mFlipSelector = 0;

	mFrameWidth = (int32) params.frameBounds.x + 1;
	mFrameHeight = (int32) params.frameBounds.y + 1;

	int32 maxBytesPerRow = __highest_supported_bitdepth() * mFrameWidth / 8;

	mBackBits = (uint8 *) ::os_malloc(mRingSize * mFrameHeight * maxBytesPerRow);
	if (NULL == mBackBits)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	mRing = new(std::nothrow) BufferRecord [mRingSize];
	if (NULL == mRing)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	for (int i = 0; i < mRingSize; i++)
	{
		mRing[i].writeLock = ::kreate_sem(0, NULL);
		mRing[i].unlockAtConnect = true;

		if (B_OK > mRing[i].writeLock)
		{
			error_text = "failure providing ring locking mechanism at framework construction.";
			error = mRing[i].writeLock;
			break;
		}

		mRing[i].bits = mBackBits + i * mFrameHeight * maxBytesPerRow;
	}

	if (B_OK != error)
		return;

	BView *transp_back = new(std::nothrow) BView(Bounds(), NULL, B_FOLLOW_ALL, B_WILL_DRAW);
	if (NULL == transp_back)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	transp_back->SetViewColor(B_TRANSPARENT_COLOR);
	AddChild(transp_back);

	if (SupportsWindowMode())
	{
		BScreen *screen = new(std::nothrow) BScreen(this);
		if (NULL == screen)
		{
			error_text = "'bad_alloc' at framework construction.";
			error = B_ERROR;
			return;
		}

		BRect scrRect = screen->Frame();
		MoveTo((scrRect.Width() - params.frameBounds.x) * .5f, (scrRect.Height() - params.frameBounds.y) * .5f);
		delete screen;
	}
	else
		SetFullScreen(true);
}

FrmDirectWindowLC::~FrmDirectWindowLC()
{
	if (NULL != mRing)
		for (int i = 0; i < mRingSize; i++)
			::delete_sem(mRing[i].writeLock);

	int32 result; 
	thread_id painter = ::find_thread(Framework::__painter_name());

	if (B_NAME_NOT_FOUND != painter)
		::wait_for_thread(painter, &result);

	delete [] mRing;
	::os_free(mBackBits);
}

void FrmDirectWindowLC::MessageReceived(BMessage *message) 
{
	switch (message->what)
	{
	case MSG_FLIP:

		mFlipLock.Lock();

		if (mConnected)
		{
			uint32 bytesPerPixel = mFrontBitsInfo.bitsPerPixel / 8;
			// note: only byte-multiple bitdepths are handled

			for (uint32 i = 0; i < mFrontClipInfo.clipRectsCount; i++)
			{
				clipping_rect &clipRect = mFrontClipInfo.clipRects[i];

				uint8 *dstBits = mFrontBits +
									clipRect.top * mFrontBitsInfo.bytesPerRow +
									clipRect.left * bytesPerPixel;

				uint8 *srcBits = mRing[mFlipSelector].bits +
									(clipRect.top - mFrontClipInfo.windowBounds.top) * mBackBitsInfo.bytesPerRow +
									(clipRect.left - mFrontClipInfo.windowBounds.left) * bytesPerPixel;

				int32 clipRectHeight = clipRect.bottom - clipRect.top + 1;
				int32 clipRectWidth	 = clipRect.right - clipRect.left + 1;

				for (int y = 0; y < clipRectHeight; y++)
				{
					::memblt(srcBits, dstBits, clipRectWidth * bytesPerPixel);

					dstBits += mFrontBitsInfo.bytesPerRow;
					srcBits += mBackBitsInfo.bytesPerRow;
				}
			}

			::release_sem(mRing[mFlipSelector].writeLock);
		}
		else
			mRing[mFlipSelector].unlockAtConnect = true;

		mFlipLock.Unlock();

		++mFlipSelector %= mRingSize; 
		break;

	default:
		BDirectWindow::MessageReceived(message);
	} 
}

bool FrmDirectWindowLC::QuitRequested()
{
	Hide(); 
	Sync();
	
 	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);

	return BDirectWindow::QuitRequested();
}

void FrmDirectWindowLC::DirectConnected(direct_buffer_info *info)
{
	mFlipLock.Lock();
	// since DirectConnected() is not a regular message hook but an app_server
	// callback, it runs asynchronously to the window looper, hence this lock
	// for synchronization with the flipping.

	switch(info->buffer_state & B_DIRECT_MODE_MASK)
	{
		case B_DIRECT_START: 

			mFrontBits = (uint8 *) info->bits;

			mFrontBitsInfo.bytesPerRow	= info->bytes_per_row;
			mFrontBitsInfo.bitsPerPixel	= info->bits_per_pixel;
			mFrontBitsInfo.pixelFormat	= info->pixel_format;

			mFrontClipInfo.windowBounds = info->window_bounds;
			mFrontClipInfo.clipRects	= NULL;

			mBackBitsInfo.bytesPerRow = info->bits_per_pixel * mFrameWidth / 8;
			// note: only byte-multiple bitdepths are handled

			for (int i = 0; i < mRingSize; i++)
				if (mRing[i].unlockAtConnect)
				{
					mRing[i].unlockAtConnect = false;
					::release_sem_etc(mRing[i].writeLock, 1, B_DO_NOT_RESCHEDULE);
				}

			mConnected = true; 

		case B_DIRECT_MODIFY: 

			mFrontClipInfo.clipRectsCount = info->clip_list_count;
			mFrontClipInfo.clipRects = (clipping_rect *) ::realloc(mFrontClipInfo.clipRects, mFrontClipInfo.clipRectsCount * sizeof(clipping_rect));
			::memcpy(mFrontClipInfo.clipRects, info->clip_list, mFrontClipInfo.clipRectsCount * sizeof(clipping_rect));

			break;

		case B_DIRECT_STOP:

			::free(mFrontClipInfo.clipRects);

			mConnected = false; 
			break; 
	}

	mFlipLock.Unlock();
}

status_t FrmDirectWindowLC::frame_open(Buffer &desc)
{
	status_t res = ::acquire_sem(mRing[mDrawSelector].writeLock);

	if (B_OK == res)
	{
		desc.frameHeight	= mFrameHeight;
		desc.frameWidth		= mFrameWidth;
		desc.pixelFormat	= mFrontBitsInfo.pixelFormat;
		desc.rowStride		= mBackBitsInfo.bytesPerRow;
		desc.bits			= mRing[mDrawSelector].bits;
	}

	return res;
}

void FrmDirectWindowLC::frame_close()
{
	PostMessage(MSG_FLIP);

	++mDrawSelector %= mRingSize; 
}
