#ifndef	__FRMHELPER_H__
#define	__FRMHELPER_H__

#include <OS.h>

#include <string.h>

#include <iostream>

// kreate_sem()		: creates a semaphore
//		- thread_count,		uint32			: initial semaphore count,	input
//		- name,				const char *	: semaphore's name,			input
// returns
//		sem_id

inline sem_id kreate_sem(uint32	thread_count, const char *name)
{
	sem_id result = ::create_sem(thread_count, name);

	if (B_OK > result)
		std::cerr << "kreate_sem(): failure at create_sem(), error code 0x" << std::hex << (int) result
				  << ", translating as \"" << ::strerror((int) result) << '\"' << std::endl;

	return result;
}

// kreate_port()	: creates a port
//		- queue_length,		int32			: port's queue length,	input
//		- name,				const char *	: port's name,			input
// returns
//		port_id

inline port_id kreate_port(int32 queue_length, const char *name)
{
	port_id result = ::create_port(queue_length, name);

	if (B_OK > result)
		std::cerr << "kreate_port(): failure at create_port(), error code 0x" << std::hex << (int) result
				  << ", translating as \"" << ::strerror((int) result) << '\"' << std::endl;

	return result;
}

// os_malloc()	: allocates the requested amount of memory by VMM pages
//		- size,			size_t		: amount of memory,			input
// returns
//		void *
// note
//		- function uses create_area() from the KernelKit
//		- memory allocated by this function must be de-allocated by os_free()

inline void *os_malloc(size_t size)
{
	void *ptr;
	area_id idArea = ::create_area("os_malloc_pages", &ptr, B_ANY_ADDRESS,
									(size + B_PAGE_SIZE - 1) & ((uint32) -1 ^ (B_PAGE_SIZE - 1)),
									B_NO_LOCK, B_READ_AREA | B_WRITE_AREA);
	if (B_ERROR < idArea)
		return ptr;

	std::cerr << "os_malloc(): failure at create_area(), error code 0x" << std::hex << (int) idArea
			  << ", translating as \"" << ::strerror((int) idArea) << '\"' << std::endl;

	return NULL;
}

// os_free()	: de-allocates memory that has been allocated through os_malloc()
//		- ptr,			void *		: memory to be de-allocated,	input
// returns
//		nil
// note
//		- function uses area_for() and delete_ared() from the KernelKit

inline void os_free(void *ptr)
{
	area_id idArea = ::area_for(ptr);
	
	if (B_ERROR < idArea)
		::delete_area(idArea);
	else
		std::cerr << "os_free(): failure at area_for(), error code 0x" << std::hex << (int) idArea
				  << ", translating as \"" << ::strerror((int) idArea) << '\"' << std::endl;
}

//////////////////////////////////////////////////////////////////////////////////////////
// class ColorRGBA	: an rgb_color wrapper providing initialization at construcition

class ColorRGBA : public rgb_color
{
public:
	ColorRGBA(uint8 arg_r, uint8 arg_g, uint8 arg_b, uint8 arg_a = 255)
	{
		red   = arg_r;
		green = arg_g;
		blue  = arg_b;
		alpha = arg_a;
	}
};

//////////////////////////////////////////////////////////////////////////////////////////
// class Lifespan	: an object lifespan watch

class Lifespan
{
public:
	typedef	sem_id	Token;

protected:
	Token	mToken;

public:
	Lifespan();
	~Lifespan();

	Token			AcquireToken();			// acquires a token for 'this' object
	static status_t WaitForDeath(Token);	// blocks until object w/ specified token gets destroyed

	// note: once AcquireToken() was successfully invoked for an object,
	// it must be succeeded by a respective WaitForDeath()
};

inline Lifespan::Lifespan() :
	mToken(B_ERROR)
{}

inline Lifespan::~Lifespan()
{
	if (B_ERROR < mToken)
		::release_sem(mToken);
}

inline Lifespan::Token Lifespan::AcquireToken()
{
	if (B_OK > mToken)
		mToken = ::kreate_sem(0, NULL);

	return mToken;
}

inline status_t Lifespan::WaitForDeath(Token token)
{
	status_t acq_res = ::acquire_sem(token);

	if (B_OK == acq_res)
		::delete_sem(token);

	return acq_res;			
}

#endif // __FRMHELPER_H__
