#include "frmWindowScreen.h"

#include <Application.h>
#include <memory.h>

#include <new>

#include "frmHelper.h"

#define	MSG_FLIP				'FLIP'

#define SANE_FLIP_PERIOD		18000	// usec, ~55.5Hz

// window ///////////////////////////////////////////////////////////////////////

uint32 FrmWindowScreen::MatchScreenSpace(BPoint &frameBounds, color_space color)
{
	static const int32 nSupportedResolutions = 6;
	static const int32 nSupportedBitDepths = 4;

	static const struct ScreenSpaceMapper
	{
		int32 width;
		int32 height;
		uint32 screen_space[nSupportedBitDepths];
	}
	mapper[nSupportedResolutions] = // keep in order with resolutions priority 
	{
		{  640,  480, { B_8_BIT_640x480,	B_15_BIT_640x480,	B_16_BIT_640x480,	B_32_BIT_640x480 }},
		{  800,  600, { B_8_BIT_800x600,	B_15_BIT_800x600,	B_16_BIT_800x600,	B_32_BIT_800x600 }},
		{ 1024,  768, { B_8_BIT_1024x768,	B_15_BIT_1024x768,	B_16_BIT_1024x768,	B_32_BIT_1024x768 }},
		{ 1152,  900, { B_8_BIT_1152x900,	B_15_BIT_1152x900,	B_16_BIT_1152x900,	B_32_BIT_1152x900 }},
		{ 1280, 1024, { B_8_BIT_1280x1024,	B_15_BIT_1280x1024,	B_16_BIT_1280x1024,	B_32_BIT_1280x1024 }},
		{ 1600, 1200, { B_8_BIT_1600x1200,	B_15_BIT_1600x1200,	B_16_BIT_1600x1200,	B_32_BIT_1600x1200 }}
	}; 

	for (int32 i = 0; i < nSupportedResolutions; i++)
	{
		if (((int32) frameBounds.x) < mapper[i].width &&
			((int32) frameBounds.y) < mapper[i].height)
		{
			switch(color)
			{
			case B_CMAP8:
			case B_GRAY8:

				return mapper[i].screen_space[0];

			case B_RGB15:
			case B_RGBA15:
			case B_RGB15_BIG:
			case B_RGBA15_BIG:

				return mapper[i].screen_space[1];

			case B_RGB16:
			case B_RGB16_BIG:

				return mapper[i].screen_space[2];

			case B_RGB32:
			case B_RGBA32:
			case B_RGB32_BIG:
			case B_RGBA32_BIG:

				return mapper[i].screen_space[3];

			default:
				return 0;
			}
		}
	}
	return 0;
}

FrmWindowScreen::FrmWindowScreen(FrameworkParams &params, status_t &error, const char * &error_text) :

	BWindowScreen("BWindowScreen Framework", MatchScreenSpace(params.frameBounds, params.pixelFormat), &error),

	mBScreen(NULL),
	mRing(NULL),
	mParams(params),
	mError(error),
	mErrorText(error_text)
{
	if (B_OK != error)
	{
		error_text = "failure constructing a BWindowScreen.";
		return;
	}

	mConnected = false;
	mDrawSelector = 0;
	mFlipSelector = 0;
	mFlipTime = (bigtime_t) 0;

	switch (MatchScreenSpace(params.frameBounds, params.pixelFormat))
	{
		case B_8_BIT_640x480:
		case B_15_BIT_640x480:
		case B_16_BIT_640x480:
		case B_32_BIT_640x480:

			mScreenSize.x = 640;
			mScreenSize.y = 480;
			break;

		case B_8_BIT_800x600:
		case B_15_BIT_800x600:
		case B_16_BIT_800x600:
		case B_32_BIT_800x600:

			mScreenSize.x = 800;
			mScreenSize.y = 600;
			break;

		case B_8_BIT_1024x768:
		case B_15_BIT_1024x768:
		case B_16_BIT_1024x768:
		case B_32_BIT_1024x768:

			mScreenSize.x = 1024;
			mScreenSize.y = 768;
			break;

		case B_8_BIT_1152x900:
		case B_15_BIT_1152x900:
		case B_16_BIT_1152x900:
		case B_32_BIT_1152x900:

			mScreenSize.x = 1152;
			mScreenSize.y = 900;
			break;

		case B_8_BIT_1280x1024:
		case B_15_BIT_1280x1024:
		case B_16_BIT_1280x1024:
		case B_32_BIT_1280x1024:

			mScreenSize.x = 1280;
			mScreenSize.y = 1024;
			break;

		case B_8_BIT_1600x1200:
		case B_15_BIT_1600x1200:
		case B_16_BIT_1600x1200:
		case B_32_BIT_1600x1200:

			mScreenSize.x = 1600;
			mScreenSize.y = 1200;
			break;

		default:

			error_text = "failure matching a screen space at framework construction.";
			error = B_ERROR;
			return;
	}

	mFrameSize.x = (int32) params.frameBounds.x + 1;
	mFrameSize.y = (int32) params.frameBounds.y + 1;

	mFramePos.x = (mScreenSize.x - mFrameSize.x) / 2;
	mFramePos.y = (mScreenSize.y - mFrameSize.y) / 2;

	mRing = new(std::nothrow) BufferRecord [params.numRingBuffers];
	if (NULL == mRing)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
		return;
	}

	for (uint32 i = 0; i < params.numRingBuffers; i++)
	{
		mRing[i].writeLock = ::kreate_sem(0, NULL);
		mRing[i].unlockAtConnect = true;

		if (B_OK > mRing[i].writeLock)
		{
			error_text = "failure providing a ring locking mechanism at framework construction";
			error = mRing[i].writeLock;
			break;
		}
	}

	if (B_OK != error)
		return;

	mBScreen = new(std::nothrow) BScreen(this);
	if (NULL == mBScreen)
	{
		error_text = "'bad_alloc' at framework construction.";
		error = B_ERROR;
	}
}

FrmWindowScreen::~FrmWindowScreen()
{
	if (NULL != mRing)
		for (uint32 i = 0; i < mParams.numRingBuffers; i++)
			::delete_sem(mRing[i].writeLock);

	int32 result; 
	thread_id painter = ::find_thread(Framework::__painter_name());

	if (B_NAME_NOT_FOUND != painter)
		::wait_for_thread(painter, &result);

	delete [] mRing;
	delete mBScreen;
}

void FrmWindowScreen::MessageReceived(BMessage *message) 
{
	switch (message->what)
	{
	case MSG_FLIP:

		if (mConnected)
		{
			if (CanControlFrameBuffer())
			{
				// unwanted things would happen if we rendered pictures at
				// a higher rate than the card can flip. we here make sure
				// not to flip more than 55.5 pictures per second if the
				// card does not support retrace syncing

		      	if (B_OK != mBScreen->WaitForRetrace())
		      	{
					bigtime_t currTime = ::system_time();

					if (currTime < mFlipTime)
					{
						::snooze(mFlipTime - currTime);
						mFlipTime += SANE_FLIP_PERIOD;
					}
					else
						mFlipTime = currTime + SANE_FLIP_PERIOD;
				}		

				MoveDisplayArea(0, mFlipSelector * mScreenSize.y);
			}

			::release_sem(mRing[mFlipSelector].writeLock); 
		}
		else
			mRing[mFlipSelector].unlockAtConnect = true;

		++mFlipSelector %= mParams.numRingBuffers; 
		break;

	default:
		BWindowScreen::MessageReceived(message);
	} 
}

bool FrmWindowScreen::QuitRequested()
{
	Hide(); 
	Sync();
	
	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);

	return BWindowScreen::QuitRequested();
}

void FrmWindowScreen::ScreenConnected(bool connected)
{
	// ScreenConnected() is a regular message hook which runs synchronously to
	// the window looper, hence it is inherently synchronized with the flipping.
	// yet explicit synchronization with the drawing code is needed as drawing
	// takes place right in video memory.

	mDrawLock.Lock();

	if (connected)
		if (CanControlFrameBuffer() && B_OK == SetFrameBuffer(mScreenSize.x, mScreenSize.y * mParams.numRingBuffers))
		{
			graphics_card_info *cardInfo = CardInfo();

			mBitsInfo.bytesPerRow = cardInfo->bytes_per_row;  
			mBitsInfo.bitsPerPixel = cardInfo->bits_per_pixel;

			// initial video buffer cleanup
			::memset(cardInfo->frame_buffer, 0, cardInfo->height * cardInfo->bytes_per_row);

			for (uint32 i = 0; i < mParams.numRingBuffers; i++)
			{
				mRing[i].bits = (uint8 *)cardInfo->frame_buffer + cardInfo->bytes_per_row * mScreenSize.y * i +
											mFramePos.y * cardInfo->bytes_per_row +
											mFramePos.x * cardInfo->bits_per_pixel / 8;

				if (mRing[i].unlockAtConnect)
				{
					mRing[i].unlockAtConnect = false;
					::release_sem_etc(mRing[i].writeLock, 1, B_DO_NOT_RESCHEDULE);
				}
			}
		}
		else
		{
			mErrorText = "cannot control frame buffer or number of ring buffers too big for video memory.";
			mError = B_ERROR;

			::be_app_messenger.SendMessage(B_QUIT_REQUESTED);
		}

	mConnected = connected;

	mDrawLock.Unlock();
}

status_t FrmWindowScreen::frame_open(Buffer &desc)
{
	status_t res;

	while (B_OK == (res = ::acquire_sem(mRing[mDrawSelector].writeLock)))
	{
		desc.frameHeight	= mFrameSize.y;
		desc.frameWidth		= mFrameSize.x;
		desc.pixelFormat 	= mParams.pixelFormat;
		desc.rowStride		= mBitsInfo.bytesPerRow;
		desc.bits			= mRing[mDrawSelector].bits;

		mDrawLock.Lock();

		if (mConnected)
			break;
	
		// acquired buffer is not actually present, try again
		mRing[mDrawSelector].unlockAtConnect = true;
		mDrawLock.Unlock();
	}

	return res; 
}

void FrmWindowScreen::frame_close()
{
	mDrawLock.Unlock();

	// frame drawn, enqueue it for displaying and advance the draw selector
	PostMessage(MSG_FLIP);
	++mDrawSelector %= mParams.numRingBuffers; 
}
