//////////////////////////////////////////////////////////////////////////////////
// app.cpp
//
// a sample application that utilizes a frame-buffer framework for the purposes
// of demonstrating THURP
//////////////////////////////////////////////////////////////////////////////////

#include <OS.h>
#include <Application.h>
#include <Screen.h>
#include <File.h>
#include <Entry.h>
#include <TranslatorRoster.h>
#include <BitmapStream.h>
#include <String.h>
#include <Roster.h>
#include <Resources.h>

#include <stdio.h>
#include <memory.h>

#include <new>
#include <iostream>
#include <vector>
#include <set>
#include <map>

#include "app_params.h"
#include "setupWindow.h"

#include "frmNull.h"
#include "frmBitmap.h"
#include "frmDirectWindow.h"
#include "frmDirectWindowLC.h"
#include "frmWindowScreen.h"
#include "frmHelper.h"

#include "rend.h"
#include "rendHelper.h"

#define	DEFAULT_NUM_RING_BUFFERS	2
#define DEFAULT_FRAMEWORK_TYPE		FRAMEWORK_BVIEW
#define DEFAULT_SAMPLING			RENDER_STATE_SAMP_CONST_NEAREST

#define CONTENT_AREA_DIM_X			512
#define CONTENT_AREA_DIM_Y			384

#define	TEST_BUMPMAPPING			false

// application sub-class(es) /////////////////////////////////////////////////////

class App : BApplication
{
	status_t				mAppStatus;
	const char *			mErrorText;
	BString					mErrorCompositeText;

	Framework *				mFramework;
	AppParams				mParams;

	bool					mSkipSetup;
	bool					mDumpStats;

	static int32			DrawingThread(void *data);
	thread_id				mDrawThreadID; 

	bigtime_t				mElapsedTime;
	uint32					mFrameCount;

	BBitmap *				mNormalMap;
	std::vector<BBitmap *>	mBaseMipmap[2];
	volatile int32			mActiveBase;

	uint32					mTexRepeat;

	inline void				WipeMipmap(int32 i);

public:
	App(status_t&);
	~App();

	virtual void			ArgvReceived(int32 argc, char **argv);		
	virtual void			RefsReceived(BMessage *message);
	virtual void			ReadyToRun();
	virtual bool			QuitRequested();

private:
	void					CreateDrawingThread_BWindow();
};

App::App(status_t& error) :
	BApplication("application/x-blu-thurpdemo", &error),

	mErrorText(NULL),
	mFramework(NULL),
	mNormalMap(NULL)
{
	if (B_OK != (mAppStatus = error))
	{
		mErrorText = "failure at BApplication construction.";
		return;
	}

	mSkipSetup = false;
	mDumpStats = false;

	mParams.frameworkType					= DEFAULT_FRAMEWORK_TYPE;
	mParams.frameworkParams.numRingBuffers	= DEFAULT_NUM_RING_BUFFERS;
	mParams.frameworkParams.pixelFormat		= B_NO_COLOR_SPACE;
	mParams.frameworkParams.frameBounds		= BPoint(CONTENT_AREA_DIM_X - 1, CONTENT_AREA_DIM_Y - 1);

	mParams.rendSampling					= DEFAULT_SAMPLING;

	mFrameCount = 0;

	mActiveBase = 0;
	mTexRepeat = 4;
}

App::~App()
{
	delete mNormalMap;
	WipeMipmap(0);
	WipeMipmap(1);		// wipe out both the active and inactive base texture mipmaps

	if (B_OK == mAppStatus)
	{
		if (mDumpStats)
		{
			uint32 i = 0;
			while ((i < AppParams::nPixFormats) && (AppParams::pixFmtId[i].space != mParams.frameworkParams.pixelFormat))
				i++;

			std::cout << "number of ring buffers " << mParams.frameworkParams.numRingBuffers << std::endl
					  << "pixel format " << AppParams::pixFmtId[i].textual << std::endl
					  << "elapsed time (sec) " << ((double)mElapsedTime / 1000000.0) << std::endl
					  << "drawn frames " << mFrameCount << std::endl
					  << "framerate (fps) " << ((mFrameCount / (double)mElapsedTime) * 1000000.0) << std::endl;
		}
	}
	else
		std::cerr << mErrorText << std::endl;
}

void App::ArgvReceived(int32 argc, char **argv)
{
	BMessage messageRefs(B_REFS_RECEIVED);

	for (int32 i = 1; (B_OK == mAppStatus) && (i < argc); i++)
	{
		if (! ::strcmp(argv[i], PARA_STATS))
		{
			mDumpStats = true;
			continue;
		}

		if (! ::strcmp(argv[i], PARA_NOSETUP))
		{
			mSkipSetup = true;
			continue;
		}

		if (! ::strcmp(argv[i], PARA_FRAMEWORK))
		{
			if (argc > ++i)
			{
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_NULL))
				{
					mParams.frameworkType = FRAMEWORK_NULL;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BVIEW))
				{
					mParams.frameworkType = FRAMEWORK_BVIEW;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BDWIN))
				{
					mParams.frameworkType = FRAMEWORK_BDIRECTWINDOW;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BDWINLC))
				{
					mParams.frameworkType = FRAMEWORK_BDIRECTWINDOW_LC;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_FRAMEWORK_BWINSCR))
				{
					mParams.frameworkType = FRAMEWORK_BWINDOWSCREEN;
					continue;
				}
			}
		}

		if (! ::strcmp(argv[i], PARA_NUMBUFFS))
		{
			if (argc > ++i)
			{
				if (1 == ::sscanf(argv[i], "%ld", &mParams.frameworkParams.numRingBuffers) &&
					MIN_NUM_RING_BUFFERS <= mParams.frameworkParams.numRingBuffers &&
					MAX_NUM_RING_BUFFERS >= mParams.frameworkParams.numRingBuffers)
					continue;
			}
		}

		if (! ::strcmp(argv[i], PARA_TEXTUREFILE)) 
		{
			if (argc > ++i)
			{
				entry_ref ref;

				BEntry entry(argv[i]);
				entry.GetRef(&ref);

				messageRefs.AddRef("refs", &ref);

				mTexRepeat = 1;
				continue;
			}
		}

		if (! ::strcmp(argv[i], PARA_SAMPLING))
		{
			if (argc > ++i)
			{
				if (! ::strcmp(argv[i], PARA_SAMPLING_CONST_NEAREST))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_CONST_NEAREST;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_CONST_LINEAR))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_CONST_LINEAR;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_NEAREST_NEAREST))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_NEAREST_NEAREST;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_NEAREST_LINEAR))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_NEAREST_LINEAR;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_LINEAR_NEAREST))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_LINEAR_NEAREST;
					continue;
				}
				if (! ::strcmp(argv[i], PARA_SAMPLING_LINEAR_LINEAR))
				{
					mParams.rendSampling = RENDER_STATE_SAMP_LINEAR_LINEAR;
					continue;
				}
			}
		}

		// this argument is invalid, hint the user of app's command-line usage
		mErrorCompositeText << "usage: "
			<< argv[0] << " [OPTION]..\n"
			<< "OPTION\n"
			<< "\t" << PARA_NOSETUP << "\t\tbypass the setup at startup\n"
			<< "\t" << PARA_NUMBUFFS << " NUM\t\tnumber of ring buffers, domain of [" << (int32) MIN_NUM_RING_BUFFERS
			<< ", " << (int32) MAX_NUM_RING_BUFFERS << "], default is " << DEFAULT_NUM_RING_BUFFERS << "\n"
			<< "\t" << PARA_FRAMEWORK << " FWTYPE\tuse the specified framebuffer framework\n"
			<< "\t" << PARA_TEXTUREFILE << " FILE\t\tuse texture from specified file\n"
			<< "\t" << PARA_SAMPLING << " SAMPTYPE\tuse specified texture sampling\n"
			<< "\t" << PARA_STATS << "\t\t\toutput runtime statistics at successfull completion\n"
			<< "FWTYPE\n"
			<< "\t" << PARA_FRAMEWORK_NULL << "\t\t\t" << DESC_FRAMEWORK_NULL << "\n"
			<< "\t" << PARA_FRAMEWORK_BVIEW << "\t\t\t" << DESC_FRAMEWORK_BVIEW << "\n"
			<< "\t" << PARA_FRAMEWORK_BDWIN << "\t\t" << DESC_FRAMEWORK_BDWIN << "\n"
			<< "\t" << PARA_FRAMEWORK_BDWINLC << "\t\t" << DESC_FRAMEWORK_BDWINLC << "\n"
			<< "\t" << PARA_FRAMEWORK_BWINSCR << "\t\t" << DESC_FRAMEWORK_BWINSCR << "\n"
			<< "SAMPTYPE\n"
			<< "\t" << PARA_SAMPLING_CONST_NEAREST << "\t\t" << DESC_SAMPLING_CONST_NEAREST << "\n"
			<< "\t" << PARA_SAMPLING_CONST_LINEAR << "\t\t" << DESC_SAMPLING_CONST_LINEAR << "\n"
			<< "\t" << PARA_SAMPLING_NEAREST_NEAREST << "\t\t" << DESC_SAMPLING_NEAREST_NEAREST << "\n"
			<< "\t" << PARA_SAMPLING_NEAREST_LINEAR << "\t\t" << DESC_SAMPLING_NEAREST_LINEAR << "\n"
			<< "\t" << PARA_SAMPLING_LINEAR_NEAREST << "\t\t" << DESC_SAMPLING_LINEAR_NEAREST << "\n"
			<< "\t" << PARA_SAMPLING_LINEAR_LINEAR << "\t\t" << DESC_SAMPLING_LINEAR_LINEAR << "\n";

		mErrorText = mErrorCompositeText.String();
		mAppStatus = B_ERROR;
	}
	
	if (B_OK == mAppStatus)
		if (! messageRefs.IsEmpty())
			::be_app_messenger.SendMessage(&messageRefs);
}

void App::ReadyToRun(void)
{
	if (B_OK == mAppStatus)
	{
		if (mSkipSetup)	// setup interface supressed; set framework's pixel format to screen's current
		{
			BScreen *screen = new(std::nothrow) BScreen();

			if (NULL != screen && screen->IsValid())
				mParams.frameworkParams.pixelFormat = screen->ColorSpace();
			else
			{
				mErrorText = "failed to construct a BScreen; cannot get screen's current color space.";
				mAppStatus = B_ERROR;
			}

			delete screen;
		}
		else // invoke the setup interface 
		{
			SetupWindow *wndSetup = new(std::nothrow) SetupWindow(mParams);

			if (NULL != wndSetup)
			{
				Lifespan::Token token = wndSetup->AcquireToken();

				if (B_ERROR < token)
				{
					wndSetup->Show();
					mAppStatus = Lifespan::WaitForDeath(token);
				}
				else
				{
					wndSetup->LockLooper();
					wndSetup->Quit();

					mErrorText = "failed to obtain a life token for the setup window.";
					mAppStatus = (status_t) token;
				}
			}
			else
			{
				mErrorText = "'bad_alloc' at setup window construction.";
				mAppStatus = B_ERROR;
			}
		}
	}

	if (B_OK == mAppStatus) // create the display framework as requested
	{
		switch (mParams.frameworkType)
		{
		case FRAMEWORK_NULL:

			mFramework = new(std::nothrow) FrmNullWindow(mParams.frameworkParams, mAppStatus, mErrorText);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BVIEW:

			mFramework = new(std::nothrow) FrmBmpWindow(mParams.frameworkParams, mAppStatus, mErrorText);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BDIRECTWINDOW:

			mFramework = new(std::nothrow) FrmDirectWindow(mParams.frameworkParams, mAppStatus, mErrorText);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BDIRECTWINDOW_LC:

			mFramework = new(std::nothrow) FrmDirectWindowLC(mParams.frameworkParams, mAppStatus, mErrorText);

			CreateDrawingThread_BWindow();
			break;

		case FRAMEWORK_BWINDOWSCREEN:

			mFramework = new(std::nothrow) FrmWindowScreen(mParams.frameworkParams, mAppStatus, mErrorText);

			CreateDrawingThread_BWindow();
			break;
		}
	}

	if (B_OK == mAppStatus && mBaseMipmap[mActiveBase].empty()) // load default textures from app resources
	{
		app_info	appInfo;
		BFile		appFile;
		BResources	appResources;

		if (B_ERROR < be_app->GetAppInfo(&appInfo) &&
			B_ERROR < appFile.SetTo(&appInfo.ref, B_READ_ONLY) &&
			B_ERROR < appResources.SetTo(&appFile))
		{
			// base texture lod(0) (an array of CLUT8)
			const int32	 tex_dimx = 16;
			const int32	 tex_dimy = 16;
			const int8	*tex_bits;
			size_t		 tex_size;

			if (NULL != (tex_bits = (const int8 *) appResources.LoadResource('TEXB', 102, &tex_size)) &&
				tex_dimx * tex_dimy * sizeof(int8) == tex_size)
			{
				BBitmap * bitmap = new BBitmap(BRect(0, 0, tex_dimx - 1, tex_dimy - 1), B_RGB32);

				for (int32 y = 0; y < tex_dimy; y++)
					bitmap->SetBits(tex_bits + tex_dimx * y, tex_dimx * sizeof(int8), bitmap->BytesPerRow() * y, B_CMAP8);

				mBaseMipmap[mActiveBase].push_back(bitmap);
			}

			// normal map (an array of DA16)
			const int32	 norm_dimx = 16;
			const int32	 norm_dimy = 16;
			const int32	*norm_bits;
			size_t		 norm_size;

			if (NULL != (norm_bits = (const int32 *) appResources.LoadResource('TEXN', 102, &norm_size)) &&
				norm_dimx * norm_dimy * sizeof(int32) == norm_size)
			{
				mNormalMap = new BBitmap(BRect(0, 0, norm_dimx - 1, norm_dimy - 1), B_RGB32);

				for (int32 y = 0; y < norm_dimy; y++)
					::memcpy((int8 *)mNormalMap->Bits() + mNormalMap->BytesPerRow() * y, norm_bits + norm_dimx * y, norm_dimx * sizeof(int32));
					// BBitmap::SetBits() seems to be broken with B_RGBA32 pixels and discards the most significand byte of B_RGB32 pixels
			}

			// note: both the base and the normal maps will be tiled as specified in mTexRepeat.
		}
		else
		{
			mErrorText = "failed to load one of several textures from the application's resources.";
			mAppStatus = B_ERROR;
		}
	}

	if (B_OK == mAppStatus)
		::resume_thread(mDrawThreadID);
	else
		::be_app_messenger.SendMessage(B_QUIT_REQUESTED);
}

inline void	App::WipeMipmap(int32 i)
{
	while (mBaseMipmap[i].size())
	{
		delete mBaseMipmap[i].back();
		mBaseMipmap[i].pop_back();
	}
}

void App::RefsReceived(BMessage *message)
{
	if (B_OK != mAppStatus)
		return;

	int32 baseIdx = mActiveBase ^ 1;	// pick the base texture currently not in use;
	WipeMipmap(baseIdx);				// wipe it out

	entry_ref ref;
	int32 i = 0;

	while(message->FindRef("refs", i++, &ref) == B_OK) // fill up its mipmap with the received bitmaps
	{
		BEntry entry(&ref, true);

		if (! entry.Exists())
			continue;

		BFile file(&entry, B_READ_ONLY);
		BBitmapStream stream;
		BBitmap *bitmap = NULL;

		if (BTranslatorRoster::Default()->Translate(&file, NULL, NULL, &stream, B_TRANSLATOR_BITMAP) == B_OK &&
			stream.DetachBitmap(&bitmap) == B_OK)
		{
			color_space cspace = bitmap->ColorSpace();

			if (B_RGBA32 == cspace || B_RGB32 == cspace)
			{
				mBaseMipmap[baseIdx].push_back(bitmap);
			}
			else
			{
				BRect bounds(bitmap->Bounds());

				BBitmap *bitmapRgb32 = new BBitmap(bounds, B_RGBA32, true);
				BView	*converter = new BView(bounds, NULL, B_FOLLOW_NONE, 0);

				bitmapRgb32->AddChild(converter);
				converter->LockLooper();
				converter->DrawBitmap(bitmap, bounds);
				converter->UnlockLooper();
				bitmapRgb32->RemoveChild(converter);

				delete converter;
				delete bitmap;

				mBaseMipmap[baseIdx].push_back(bitmapRgb32);
			}
		}
	}

	if (mBaseMipmap[baseIdx].size()) // set the new mipmap as active
		mActiveBase = baseIdx;

	// warning:
	// we'd be in trouble if this handler got invoked more than once during a single frame (not supposed to happen)
}

bool App::QuitRequested(void)
{
	return BApplication::QuitRequested();
}

void App::CreateDrawingThread_BWindow()
{
	// try to down-cast the framework to a BWindow
	BWindow *window = dynamic_cast<BWindow *>(mFramework);

	if (NULL == window)
	{
		if (B_OK == mAppStatus)
		{
			mErrorText = "expected a BWindow-ascending framework, got something else.";
			mAppStatus = B_ERROR;
		}
		return;
	}

	if (B_OK == mAppStatus)
	{
		mDrawThreadID = ::spawn_thread(DrawingThread, Framework::__painter_name(), B_NORMAL_PRIORITY, (void *) this);

		if (B_ERROR < mDrawThreadID)
		{
			window->Show();
			return;
		}

		mErrorText = "failed to spawn a drawing thread.";
		mAppStatus = (status_t) mDrawThreadID;
	}

	// either a spawn failure or an earlier failure: destroy the framework
	window->LockLooper();
	window->Quit();
}

//////////////////////////////////////////////////////////////////////////////////
// rendition helpers
//////////////////////////////////////////////////////////////////////////////////

class color_formats_map : std::map< color_space, uint32 >
{
public:

	color_formats_map();

	uint32 fetch(color_space) const;
};

color_formats_map::color_formats_map()
{
	insert(value_type( B_RGB15,  RENDER_STATE_PIX_ARGB1555 ));
	insert(value_type( B_RGBA15, RENDER_STATE_PIX_ARGB1555 ));
	insert(value_type( B_RGB32,  RENDER_STATE_PIX_ARGB8888 ));
	insert(value_type( B_RGBA32, RENDER_STATE_PIX_ARGB8888 ));
	insert(value_type( B_RGB16,  RENDER_STATE_PIX_RGB565 ));
	insert(value_type( B_RGB24,  RENDER_STATE_PIX_RGB888 ));
}

uint32 color_formats_map::fetch(color_space space) const
{
	const_iterator it = find(space);

	if (it != end())
		return it->second;

	return RENDER_STATE_PIX_INVALID;
}

static color_formats_map cf_map;


template < typename VECT_TYPE_T, unsigned VECT_COUNT_T >
class vector_array
{
public:
	VECT_TYPE_T	(*p)[VECT_COUNT_T];
	unsigned	n;

	vector_array(VECT_TYPE_T (*argp)[VECT_COUNT_T], const unsigned argn) :
		p(argp),
		n(argn)
	{}
};


class transform
{
	float	m[4][4];

public:

	void SetIdentity()
		{ Identity(m); }

	void SetRotation(const float rad, float x, float y, float z)
		{ Rotate(m, rad, x, y, z); }

	void SetTranslation(const float x, const float y, const float z)
		{ Translate(m, x, y, z); }

	void SetScale(const float x, const float y, const float z)
		{ Scale(m, x, y, z); }
};


class surface
{
public:
	typedef	unsigned			index;

	vector_array< float, 4 >	v;		// pool of vertices
	vector_array< float, 3 >	n;		// pool of normals
	vector_array< index, 6 >	tri;	// 3 indices of vertices + 3 indices of normals
	transform					t;		// to object space (always 1st operator in a composite transform)

	surface(float (*verts)[4], const unsigned nverts,
			float (*norms)[3], const unsigned nnorms,
			index (*tris)[6] = NULL, const unsigned ntris = 0) :
		v(verts, nverts),
		n(norms, nnorms),
		tri(tris, ntris)
	{}

	static surface *arr_construct(float (*)[4], const unsigned,
								  float (*)[3], const unsigned,
								  const unsigned);

	static void	arr_destroy(surface *);
};

surface *surface::arr_construct(float (*verts)[4], const unsigned nverts,
								float (*norms)[3], const unsigned nnorms,
								const unsigned count)
{
	surface *p = (surface *) ::malloc(sizeof(surface) * count);

	if (NULL != p)
		for (unsigned i = 0; i < count; i++)
			new (p + i) surface(verts, nverts, norms, nnorms);
	
	return p;
}

void surface::arr_destroy(surface *p)
{
	::free(p);
}


static status_t bail_out(const char *err_msg, const status_t err_code)
{
	std::cerr << err_msg << std::endl;

	::be_app_messenger.SendMessage(B_QUIT_REQUESTED);
	return err_code;
}


int32 App::DrawingThread(void *data)
{
	if (NULL == data)
	{
		return bail_out("drawing thread err: invalid argument", B_ERROR);
	}

	App &app = *(App *)data;

	if (NULL == app.mFramework)
	{
		return bail_out("drawing thread err: no framework", B_ERROR);
	}

	Framework &frmwrk = *app.mFramework;

	// define some object to be displayed;
	// let that be a cube composed through six orthogonal rotations of a square surface
	static const unsigned vcount = 4;
	static /*const*/ float v[vcount][4] =
	{
		{ -1,  1,  1,  1  },
		{  1,  1,  1,  1  },
		{ -1, -1,  1,  1  },
		{  1, -1,  1,  1  }
	};

	static const unsigned ncount = 1;
	static /*const*/ float n[ncount][3] =
	{
		{  0,  0,  1  }
	};

	static /*const*/ float uv[vcount][2];

	static const unsigned tricount = 2;
	static /*const*/ surface::index square[tricount][6] =
	{
		{  0,  1,  2,  0,  0,  0  },
		{  2,  1,  3,  0,  0,  0  }
	};

	// six surfaces to define a cube
	static const unsigned scount = 6;
	surface * cube = surface::arr_construct(v, vcount, n, ncount, scount);
	
	if (NULL == cube)
	{
		return bail_out("drawing thread err: failed to allocate memory for the object definition", B_ERROR);
	}

	cube[0].tri.p = square;
	cube[0].tri.n = tricount;
	cube[0].t.SetIdentity();					// front side

	cube[1].tri.p = square;
	cube[1].tri.n = tricount;
	cube[1].t.SetRotation(	M_PI,	0, 1, 0);	// back side

	cube[2].tri.p = square;
	cube[2].tri.n = tricount;
	cube[2].t.SetRotation( -M_PI_2, 0, 1, 0);	// left side

	cube[3].tri.p = square;
	cube[3].tri.n = tricount;
	cube[3].t.SetRotation(	M_PI_2, 0, 1, 0);	// right side

	cube[4].tri.p = square;
	cube[4].tri.n = tricount;
	cube[4].t.SetRotation( -M_PI_2, 1, 0, 0);	// ceil side

	cube[5].tri.p = square;
	cube[5].tri.n = tricount;
	cube[5].t.SetRotation(	M_PI_2, 1, 0, 0);	// floor side

	// obtain an aux buffer of adequate bitdepth (for multi-pass) and of max dimensions (to avoid buffer resizing)
	static const unsigned auxPixFormat	= RENDER_STATE_PIX_RGB888;
	static const unsigned auxPixSize	= PixelRGB888::PackedSize();
	static const unsigned auxHeight		= CONTENT_AREA_DIM_Y;
	static const unsigned auxWidth		= CONTENT_AREA_DIM_X;

	rend_state.aux.row_stride	= auxWidth * auxPixSize;
	rend_state.aux.pix_format	= auxPixFormat;
	rend_state.aux.bits			= (int8 *) ::os_malloc(auxHeight * auxWidth * auxPixSize);

	span_above.column = (float (*)[RENDER_STATE_SPANBUFF_VECTSIZE])
		::os_malloc((auxWidth + 1) * sizeof(float [RENDER_STATE_SPANBUFF_VECTSIZE]));

	if (NULL == rend_state.aux.bits || NULL == span_above.column)
	{
		return bail_out("drawing thread err: failed to allocate memory for the aux and/or the span buffer", B_ERROR);
	}

	// pre-set the scene lighting and object material; both material and light source defined by the opengl canon
	static const unsigned lscount = 1;
	LightSource source[lscount];

	source[0].specular = Color(1, 1, 1, 1);

	Illumination lighting;

	lighting.num_src	  = lscount;
	lighting.src		  = source;
	lighting.local_viewer = true;

	Material mater;

	mater.diffuse	= Color(1,  .3, .5, 1);
	mater.specular	= Color(1,  1,  1,  1);
	mater.shininess	= 6;

	// the animation parameters
	float angle = 0;
	static const float ang_inc = 2.f * M_PI / 200.f;

	// the animation heartbeat - the frame loop
	Framework::Buffer desc;
	bigtime_t t0 = ::system_time();

	while (B_OK == frmwrk.frame_open(desc))
	{
		size_t pixel_chunk;
		size_t pixels_per_chunk;
		size_t row_alignment;

		::get_pixel_size_for(desc.pixelFormat, &pixel_chunk, &row_alignment, &pixels_per_chunk);

		int32 loop = desc.frameHeight;
		int32 bytes_per_row = (desc.frameWidth * pixel_chunk) / pixels_per_chunk;

		for (int8 *row = (int8 *) desc.bits; loop--; row += desc.rowStride)
			::memset(row, 0x80, bytes_per_row);

		// setup the render state
		viewport.screen.left	= 0;
		viewport.screen.top		= 0;
		viewport.screen.right	= desc.frameWidth;
		viewport.screen.bottom	= desc.frameHeight;

		viewport.depth_range_n	= 0;
		viewport.depth_range_f	= 1;

		rend_state.frame.pix_format	= cf_map.fetch(desc.pixelFormat);
		rend_state.frame.row_stride	= desc.rowStride;
		rend_state.frame.bits		= (int8 *) desc.bits;

		rend_state.culling	= RENDER_STATE_CULL_CCW;
		rend_state.sampling	= app.mParams.rendSampling;

		if (TEST_BUMPMAPPING)
		{
			const BBitmap * normmap	= app.mNormalMap;
			
			rend_state.tex.bits			= (int8 *) normmap->Bits();
			rend_state.tex.dimx			= normmap->Bounds().IntegerWidth() + 1;
			rend_state.tex.dimy			= normmap->Bounds().IntegerHeight() + 1;
			rend_state.tex.row_stride	= normmap->BytesPerRow();
			rend_state.tex.pix_format	= RENDER_STATE_PIX_DA16;
			
			// fixup the UV mapping (as we use absolute tex coords)
			uv[1][0] = (float) (rend_state.tex.dimx * app.mTexRepeat - 1);
			uv[3][0] = (float) (rend_state.tex.dimx * app.mTexRepeat - 1);
			uv[2][1] = (float) (rend_state.tex.dimy * app.mTexRepeat - 1);
			uv[3][1] = (float) (rend_state.tex.dimy * app.mTexRepeat - 1);
		}
		else
		{
			std::vector<BBitmap *> &mipmap = app.mBaseMipmap[app.mActiveBase];
			std::vector<BBitmap *>::const_iterator ti;
			unsigned t;

			for (ti = mipmap.begin(), t = 0; (ti != mipmap.end()) && (t != RENDER_STATE_MIPMAP_CAPACITY); ti++, t++)
			{
				rend_state.tex_lod[t].bits			= (int8 *) (**ti).Bits();
				rend_state.tex_lod[t].dimx			= (**ti).Bounds().IntegerWidth() + 1;
				rend_state.tex_lod[t].dimy			= (**ti).Bounds().IntegerHeight() + 1;
				rend_state.tex_lod[t].row_stride	= (**ti).BytesPerRow();
				rend_state.tex_lod[t].pix_format	= cf_map.fetch((**ti).ColorSpace());
			}

			rend_state.tex_lod_min = 0;
			rend_state.tex_lod_max = t - 1;

			// fixup the UV mapping (as we use absolute tex coords)
			uv[1][0] = (float) (rend_state.tex_lod[0].dimx * app.mTexRepeat - 1);
			uv[3][0] = (float) (rend_state.tex_lod[0].dimx * app.mTexRepeat - 1);
			uv[2][1] = (float) (rend_state.tex_lod[0].dimy * app.mTexRepeat - 1);
			uv[3][1] = (float) (rend_state.tex_lod[0].dimy * app.mTexRepeat - 1);
		}

		// prepare the object & scene transformations that will be used
		float t1[4][4];
		Rotate(t1, angle, 1, 1, -1);

		angle = ::fmodf(angle + ang_inc, 2.f * M_PI);

		static const float sqrt3 = ::sqrtf(3);

		float t2[4][4];
		Translate(t2, 0, 0, -(1.f + sqrt3));

		float t3[4][4];
		Frustum(t3, -1, 1, -1, 1, 1, 5);

		// render the object iterating by its surfaces
		for (unsigned si = 0; si < scount; si++)
		{
			float t[4][4], mvI[4][4], mvIT[4][4];

			// produce the model-view transformation as t2 * t1 * t_surface
			MulMat4(t1, (float(&)[4][4])cube[si].t, t);
			MulMat4(t2, t);

			Color diff[vcount];
			Color spec[vcount];

			if (InvertMat4(t, mvI))
			{
				if (TEST_BUMPMAPPING)
				{
					// in our case all surface vertices share one tangent space, which coincides with model space
					vect<3> ldir = - vect<3>::adopt(lighting.src[0].spot_direction);
					vect<3> ldir_tan;

					Transform3<3>(mvI, (float (&)[3]) ldir, (float (&)[3]) ldir_tan);

					// store the resulting vector in the diffuse color
					std::set< surface::index > done;

					for (unsigned ti = 0; ti < cube[si].tri.n; ti++)
						for (unsigned i = 0; i < 3; i++)
						{
							surface::index idx = cube[si].tri.p[ti][i];
							std::pair< std::set< surface::index >::iterator, bool > res = done.insert(idx);

							if (res.second)
								*(float (*)[3])((float *)(diff + idx) + 1) = (float (&)[3]) ldir_tan;
						}
				}
				else
				{
					TranspMat4(mvI, mvIT);

					std::set< surface::index > done;

					for (unsigned ti = 0; ti < cube[si].tri.n; ti++)
						for (unsigned i = 0; i < 3; i++)
						{
							surface::index idx = cube[si].tri.p[ti][i];
							surface::index ndx = cube[si].tri.p[ti][i + 3];
							std::pair< std::set< surface::index >::iterator, bool > res = done.insert(idx);

							if (res.second)
								LightVertex(lighting, mater, t, mvIT, cube[si].v.p[idx], cube[si].n.p[ndx], diff[idx], spec[idx]);
						}
				}
			}
			else
				std::cerr << "drawing thread err: model-view matrix non-invertible" << std::endl;

			// form the complete transformation by introducing the projection
			MulMat4(t3, t);

			// take all vertices of this surface to screen space, providing enough space for the later attributes adding
			static const unsigned vectsize = 8; // xyzw + argb
			float vs[vcount][vectsize];
		
			std::set< surface::index > done;

			for (unsigned ti = 0; ti < cube[si].tri.n; ti++)
				for (unsigned i = 0; i < 3; i++)
				{
					surface::index idx = cube[si].tri.p[ti][i];
					std::pair< std::set< surface::index >::iterator, bool > res = done.insert(idx);

					if (res.second)
					{
						float vc[4];

						Transform4< 4 >(t, cube[si].v.p[idx], vc);
						// note: near/far planes clipping is ommitted, relying that the object is within those
						MapToScreen< 4 >(const_cast< const float(&)[4] >(vc), *(float (*)[4])&vs[idx]);
						// above const_cast circumvents a bug in the compiler frontend
					}
				}

			// iterate by the triangles of the surface
			for (unsigned ti = 0; ti < cube[si].tri.n; ti++)
			{
				surface::index idx0 = cube[si].tri.p[ti][0];
				surface::index idx1 = cube[si].tri.p[ti][1];
				surface::index idx2 = cube[si].tri.p[ti][2];

				// direct rendition to the max-bitdepth aux buffer by swapping it with the frame buffer
				Swap<PixelBuffer>(rend_state.frame, rend_state.aux);

				// texturing pass
				rend_state.vert_format	= RENDER_STATE_VERT_UV;
				rend_state.color_op		= RENDER_STATE_OP_COPY;

				*(float (*)[2])((float *)&vs[idx0] + 4) = (float (&)[2])(vect<2>::adopt(uv[idx0]) * vs[idx0][3]);
				*(float (*)[2])((float *)&vs[idx1] + 4) = (float (&)[2])(vect<2>::adopt(uv[idx1]) * vs[idx1][3]);
				*(float (*)[2])((float *)&vs[idx2] + 4) = (float (&)[2])(vect<2>::adopt(uv[idx2]) * vs[idx2][3]);

				PlotTriangle6(*(float (*)[6])&vs[idx0],
							  *(float (*)[6])&vs[idx1],
							  *(float (*)[6])&vs[idx2]);

				// lighting pass(es)
				rend_state.vert_format	= RENDER_STATE_VERT_ARGB;

				*(float (*)[4])((float *)&vs[idx0] + 4) = (float (&)[4])(vect<4>::adopt(*(float (*)[4])&diff[idx0]) * vs[idx0][3]);
				*(float (*)[4])((float *)&vs[idx1] + 4) = (float (&)[4])(vect<4>::adopt(*(float (*)[4])&diff[idx1]) * vs[idx1][3]);
				*(float (*)[4])((float *)&vs[idx2] + 4) = (float (&)[4])(vect<4>::adopt(*(float (*)[4])&diff[idx2]) * vs[idx2][3]);

				if (TEST_BUMPMAPPING)
					// texture is a normal map - dot-product it with the light vector in tangent space
					rend_state.color_op	= RENDER_STATE_OP_DOT3;
				else
				{
					// texture is a diffuse map - modulate it by the diffuse light
					rend_state.color_op	= RENDER_STATE_OP_MODULATE;

					PlotTriangle8(vs[idx0],
								  vs[idx1],
								  vs[idx2]);

					// add the specularity on top
					rend_state.color_op	= RENDER_STATE_OP_ADD;

					*(float (*)[4])((float *)&vs[idx0] + 4) = (float (&)[4])(vect<4>::adopt(*(float (*)[4])&spec[idx0]) * vs[idx0][3]);
					*(float (*)[4])((float *)&vs[idx1] + 4) = (float (&)[4])(vect<4>::adopt(*(float (*)[4])&spec[idx1]) * vs[idx1][3]);
					*(float (*)[4])((float *)&vs[idx2] + 4) = (float (&)[4])(vect<4>::adopt(*(float (*)[4])&spec[idx2]) * vs[idx2][3]);
				}

				PlotTriangle8(vs[idx0],
							  vs[idx1],
							  vs[idx2]);

				// swap back the frame and aux buffers
				Swap<PixelBuffer>(rend_state.frame, rend_state.aux);

				// copy what has been just rendered in the aux buffer to the frame buffer
				rend_state.vert_format	= RENDER_STATE_VERT_BASIC;
				rend_state.color_op		= RENDER_STATE_OP_COPY;

				PlotTriangle4(*(float (*)[4])&vs[idx0],
							  *(float (*)[4])&vs[idx1],
							  *(float (*)[4])&vs[idx2]);
			}
		}

		// frame completed - commit it and update the frame counter
		frmwrk.frame_close();
		app.mFrameCount++;
	}

	app.mElapsedTime = ::system_time() - t0;

	::os_free(span_above.column);
	::os_free(rend_state.aux.bits);

	surface::arr_destroy(cube);

	return B_OK;
}

// main /////////////////////////////////////////////////////////////////

int main()
{
	status_t error = B_OK;

	new(std::nothrow) App(error);

	if (NULL != ::be_app)
	{
		if (B_OK == error)	// construction OK - pass control to the app
			::be_app->Run();
		else				// something went wrong at app construction
			std::cerr << "failure at constructing the application object." << std::endl;

		delete ::be_app;
	}
	else
	{
		std::cerr << "failed to allocate memory for the application object." << std::endl;
		error = B_ERROR;
	}

	return (int) error;
}
