#ifndef __RENDCLIPFRUST__
#define __RENDCLIPFRUST__

#include "rendVect.h"

// ClipByW()		: clips a polygon so its vertices' n-th coords are greater-or-equal/less-or-equal than -W/W
//		VECTSIZE_T									: must be of R4 or greater
//		DIST_ROUTINE_T								: routine calculating the distance between a value and a limit,
//													  distance is negative if value is off the limit
//		- coord,	int32							: index of the coordinate to clip by,		input
//		- poly,		const float (**)[VECTSIZE_T]	: array containing a polygon definition,	input/output
//		- start,	uint32 &						: polygon starting vertex in the array,		input/output
//		- size,		uint32 &						: number of vertices in the polygon,		input/output
//		- buffer,	float (*)[VECTSIZE_T]			: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &						: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the polygon could be either clockwise or counter-clockwise, orientation is preserved;
//		- result may be a point or a line segment for certain marginal cases

template < int VECTSIZE_T,
		   float (DIST_ROUTINE_T)(const float, const float) >

inline void ClipByW(int32 coord,
					const float (**poly)[VECTSIZE_T],
					uint32 &start,
					uint32 &size,
					float (*buffer)[VECTSIZE_T],
					uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		float distCurr = DIST_ROUTINE_T((*poly[currVert])[coord], (*poly[currVert])[3]);
		float distPrev = DIST_ROUTINE_T((*poly[prevVert])[coord], (*poly[prevVert])[3]);

		if (distCurr < 0) // is current vertex outside?
		{
			if (distPrev > 0) // is previous vertex strictly inside?
			{
				Interpolate< VECTSIZE_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if (distCurr > 0 &&	// is current vertex strictly inside and
				distPrev < 0)	// is previous vertex outside?
			{
				Interpolate< VECTSIZE_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

inline float dist_cond_GE(const float val, const float lim)
{
	return val + lim;
}

inline float dist_cond_LE(const float val, const float lim)
{
	return lim - val;
}

// ClipByFrustum()	: clips a polygon by the view frustum
//		VECTSIZE_T									: must be of R4 or greater
//		- poly,		const float (**)[VECTSIZE_T]	: array containing a polygon definition,	input/output
//		- start,	uint32 &						: polygon starting vertex in the array,		input/output
//		- size,		uint32 &						: number of vertices in the polygon,		input/output
//		- buffer,	float (*)[VECTSIZE_T]			: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &						: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- the result is a valid polygon if its size is at least 3;
//		- size would be 0 if clipping discarded all vertices

template < int VECTSIZE_T >
inline void ClipByFrustum(const float (**poly)[VECTSIZE_T],
						  uint32 &start,
						  uint32 &size,
						  float (*buffer)[VECTSIZE_T],
						  uint32 &nbuff)
{
	// -W <= X <= W
	ClipByW< VECTSIZE_T, dist_cond_GE >(0, poly, start, size, buffer, nbuff);
	ClipByW< VECTSIZE_T, dist_cond_LE >(0, poly, start, size, buffer, nbuff);

	// -W <= Y <= W
	ClipByW< VECTSIZE_T, dist_cond_GE >(1, poly, start, size, buffer, nbuff);
	ClipByW< VECTSIZE_T, dist_cond_LE >(1, poly, start, size, buffer, nbuff);

	// -W <= Z <= W
	ClipByW< VECTSIZE_T, dist_cond_GE >(2, poly, start, size, buffer, nbuff);
	ClipByW< VECTSIZE_T, dist_cond_LE >(2, poly, start, size, buffer, nbuff);
}

#endif // __RENDCLIPFRUST__
