#ifndef __RENDHELPER__
#define __RENDHELPER__

#include <math.h>

// Swap()	: swaps two variables of the same type
//		COMMON_T					: a type to perform the swapping on
//		- a,		<COMMON_T> &	: 1st veriable,		input/output
//		- b,		<COMMON_T> &	: 2nd veriable,		input/output
// returns
//		nil
template < typename COMMON_T >
inline void Swap(COMMON_T &a, COMMON_T &b)
{
	COMMON_T tmp = a;
	a = b;
	b = tmp;
}

// Min()	: returns the lesser of two variables of the same type
//		COMMON_T					: a type to perform the comparison on
//		- a,	const <COMMON_T> &	: 1st veriable,		input
//		- b,	const <COMMON_T> &	: 2nd veriable,		input
// returns
//		const <COMMON_T> &
template < typename COMMON_T >
inline const COMMON_T& Min(const COMMON_T& a, const COMMON_T& b)
{
	return a > b ? b : a;
} 

// Max()	: returns the greater of two variables of the same type
//		COMMON_T					: a type to perform the comparison on
//		- a,	const <COMMON_T> &	: 1st veriable,		input
//		- b,	const <COMMON_T> &	: 2nd veriable,		input
// returns
//		const <COMMON_T> &
template < typename COMMON_T >
inline const COMMON_T& Max(const COMMON_T& a, const COMMON_T& b)
{
	return a < b ? b : a;
} 

// ToInt()	: performs a float-to-int conversion using the default fpu rounding mode, i.e. 'nearest'
//		- x,	float	: fp value,		input
// returns
//		long int
// note
//		- function pops the fp stack; the compiler may decide to preserve the fp stack top inserting
//		  an "fld %st(0)" before the function invokation if the passed argument is an object referred
//		  past the invokation. particularly this may happen when the argument is a non-temporary
//		  object (i.e. a declared variable). for example:
//
//		  {
//			float a, b;
//			...
//			int c = ToInt(a + b);	// "a + b" is a temporary object not used furter, dropping it off the stack is ok
//			int d = ToInt(a);		// "a" is a non-temporary object not referred any furter; dropping it is still ok
//			...						// do something with "c" & "d" here
//		  }
//
//		  {
//			float a, b;
//			...
//			int c = ToInt(a);		// "a" gets used later in "a + b", so we get an "fld %st(0)" penalty here
//			int d = ToInt(a + b);	// "a + b" is a temporary object not used any furter and can be discard freely
//			...						// do something with "c" & "d" here
//		  }
inline long int ToInt(float x)
{
	long int result;

#ifdef __INTEL__

	asm ("fistpl %0"                       
		: "=m" (result)
		: "t" (x)
		: "cc", "st");

#else

	result = (long int) x;

#endif

	return result;
}

// ToIntRound()	: performs a float-to-int conversion using the default rounding mode 'nearest'
//		- x,	float	: fp value,		input
// returns
//		long int
// note
//		- see ToInt()
inline long int ToIntRound(float x)
{
#ifdef __INTEL__

	return ToInt(x);

#elif __POWERPC__

	return lrintf(x);

#else

	#error "ToIntRound() : unsupported target architecture"

#endif
}

#endif // __RENDHELPER__
