#include "rendVect.h"
#include "rendHelper.h"
#include "rendLight.h"

inline float HalfspaceDot(const float (&v0)[3], const float (&v1)[3])
{
	float dot = vect<3>::adopt(v0) * vect<3>::adopt(v1);

	return Max(dot, 0.f);
}

// LightVertex()	: computes vertex lighting using Phong lighting model
//		- illum,		const Illumination &	: lighting set,								input
//		- surf,			const Material &		: surface material,							input
//		- mat,			const float (&)[4][4]	: model-view matrix,						input
//		- matIT,		const float (&)[4][4]	: transposed inverse of model-view matrix,	input
//		- vi,			const float (&)[4]		: vertex,									input
//		- ni,			const float (&)[3]		: vertex normal,							input
//		- color,		Color &					: resulting color							output
// returns
//		nil
// note
//		- lighting is single-sided
void LightVertex(const Illumination &illum, const Material &surf,
				 const float (&mat)[4][4], const float (&matIT)[4][4], /*const*/ float (&vi)[4], /*const*/ float (&ni)[3],
				 Color &color)
{
	float v[4];
	float n[3];

	Transform4<4>(mat,	 vi, v);
	Transform3<3>(matIT, ni, n);

	if (illum.renorm)
		vect<3>::adopt(n) *= 1.f / (vect<3>::adopt(n)).norm();

	color.a = surf.emissive.a;
	color.r = surf.emissive.r;
	color.g = surf.emissive.g;
	color.b = surf.emissive.b;

	color.a += surf.ambient.a * illum.ambient.a;
	color.r += surf.ambient.r * illum.ambient.r;
	color.g += surf.ambient.g * illum.ambient.g;
	color.b += surf.ambient.b * illum.ambient.b;
	
	for (int i = 0; i < illum.num_src; i++)
	{
		// vertex-to-source direction
		vect<3> src_dir = vect<3>::adopt(*(float (*)[3])&illum.src[i].position);

		if (illum.src[i].position[3] != 0.f)
			src_dir -= vect<3>::adopt(*(float (*)[3])&v);

		float src_norm = src_dir.norm();

		src_dir *= 1.f / src_norm;

		// spotlight factor
		float spot = 1.f;

		if (illum.src[i].spot_cutoff != (float)M_PI)
		{
			const vect<3> &neg_src_dir = -src_dir;
			float ang_cos = HalfspaceDot((float (&)[3])neg_src_dir, illum.src[i].spot_direction);

			if (ang_cos < cosf(illum.src[i].spot_cutoff))
				break;

			spot = powf(ang_cos, illum.src[i].spot_exponent);
		}

		// attenuation factor
		float att = 1.f;

		if (illum.src[i].position[3] != 0.f)
			att /= illum.src[i].attenu_constant +
				   illum.src[i].attenu_linear * src_norm +
				   illum.src[i].attenu_quadratic * src_norm * src_norm;

		// fall-off factor
		float foff = HalfspaceDot(n, (float (&)[3])src_dir);

		if (foff != 0.f) // is surface facing this light source?
		{
			// Blinn's reflection half-direction
			vect<3> hdir = src_dir;

			if (illum.local_viewer)
				hdir -= vect<3>::adopt(*(float (*)[3])&v);
			else
				hdir[2] += 1.f;

			hdir *= 1.f / hdir.norm();

			// specularity factor
			float spec = powf(HalfspaceDot(n, (float (&)[3])hdir), surf.shininess);

			color.a += att * spot *
						(surf.ambient.a  * illum.src[i].ambient.a + 
						 surf.diffuse.a  * illum.src[i].diffuse.a * foff +
						 surf.specular.a * illum.src[i].specular.a * spec);
			color.r += att * spot *
						(surf.ambient.r  * illum.src[i].ambient.r + 
						 surf.diffuse.r  * illum.src[i].diffuse.r * foff +
						 surf.specular.r * illum.src[i].specular.r * spec);
			color.g += att * spot *
						(surf.ambient.g  * illum.src[i].ambient.g + 
						 surf.diffuse.g  * illum.src[i].diffuse.g * foff +
						 surf.specular.g * illum.src[i].specular.g * spec);
			color.b += att * spot *
						(surf.ambient.b  * illum.src[i].ambient.b + 
						 surf.diffuse.b  * illum.src[i].diffuse.b * foff +
						 surf.specular.b * illum.src[i].specular.b * spec);
		}
		else // surface is facing away from this light source
		{
			color.a += att * spot * surf.ambient.a * illum.src[i].ambient.a;
			color.r += att * spot * surf.ambient.r * illum.src[i].ambient.r;
			color.g += att * spot * surf.ambient.g * illum.src[i].ambient.g;
			color.b += att * spot * surf.ambient.b * illum.src[i].ambient.b;
		}
	}
	
	color.a = Min(Max(color.a, 0.f), 1.f);
	color.r = Min(Max(color.r, 0.f), 1.f);
	color.g = Min(Max(color.g, 0.f), 1.f);
	color.b = Min(Max(color.b, 0.f), 1.f);
}

// LightVertex()	: computes vertex lighting using Phong lighting model
//		- illum,		const Illumination &	: lighting set,								input
//		- surf,			const Material &		: surface material,							input
//		- mat,			const float (&)[4][4]	: model-view matrix,						input
//		- matIT,		const float (&)[4][4]	: transposed inverse of model-view matrix,	input
//		- vi,			const float (&)[4]		: vertex,									input
//		- ni,			const float (&)[3]		: vertex normal,							input
//		- color0,		Color &					: resulting color sans specular				output
//		- color1,		Color &					: resulting specular color					output
// returns
//		nil
// note
//		- lighting is single-sided
void LightVertex(const Illumination &illum, const Material &surf,
				 const float (&mat)[4][4], const float (&matIT)[4][4], /*const*/ float (&vi)[4], /*const*/ float (&ni)[3],
				 Color &color0, Color &color1)
{
	float v[4];
	float n[3];

	Transform4<4>(mat,	 vi, v);
	Transform3<3>(matIT, ni, n);

	if (illum.renorm)
		vect<3>::adopt(n) *= 1.f / (vect<3>::adopt(n)).norm();

	color0.a = surf.emissive.a;
	color0.r = surf.emissive.r;
	color0.g = surf.emissive.g;
	color0.b = surf.emissive.b;

	color0.a += surf.ambient.a * illum.ambient.a;
	color0.r += surf.ambient.r * illum.ambient.r;
	color0.g += surf.ambient.g * illum.ambient.g;
	color0.b += surf.ambient.b * illum.ambient.b;

	color1.a = 0.f;
	color1.r = 0.f;
	color1.g = 0.f;
	color1.b = 0.f;

	for (int i = 0; i < illum.num_src; i++)
	{
		// vertex-to-source direction
		vect<3> src_dir = vect<3>::adopt(*(float (*)[3])&illum.src[i].position);

		if (illum.src[i].position[3] != 0.f)
			src_dir -= vect<3>::adopt(*(float (*)[3])&v);

		float src_norm = src_dir.norm();

		src_dir *= 1.f / src_norm;

		// spotlight factor
		float spot = 1.f;

		if (illum.src[i].spot_cutoff != (float)M_PI)
		{
			const vect<3> &neg_src_dir = -src_dir;
			float ang_cos = HalfspaceDot((float (&)[3])neg_src_dir, illum.src[i].spot_direction);

			if (ang_cos < cosf(illum.src[i].spot_cutoff))
				break;

			spot = powf(ang_cos, illum.src[i].spot_exponent);
		}

		// attenuation factor
		float att = 1.f;

		if (illum.src[i].position[3] != 0.f)
			att /= illum.src[i].attenu_constant +
				   illum.src[i].attenu_linear * src_norm +
				   illum.src[i].attenu_quadratic * src_norm * src_norm;

		// fall-off factor
		float foff = HalfspaceDot(n, (float (&)[3])src_dir);

		if (foff != 0.f) // is surface facing this light source?
		{
			// Blinn's reflection half-direction
			vect<3> hdir = src_dir;

			if (illum.local_viewer)
				hdir -= vect<3>::adopt(*(float (*)[3])&v);
			else
				hdir[2] += 1.f;

			hdir *= 1.f / hdir.norm();

			// specularity factor
			float spec = powf(HalfspaceDot(n, (float (&)[3])hdir), surf.shininess);

			// ambient & diffuse contribution
			color0.a += att * spot *
						(surf.ambient.a  * illum.src[i].ambient.a + 
						 surf.diffuse.a  * illum.src[i].diffuse.a * foff);
			color0.r += att * spot *
						(surf.ambient.r  * illum.src[i].ambient.r + 
						 surf.diffuse.r  * illum.src[i].diffuse.r * foff);
			color0.g += att * spot *
						(surf.ambient.g  * illum.src[i].ambient.g + 
						 surf.diffuse.g  * illum.src[i].diffuse.g * foff);
			color0.b += att * spot *
						(surf.ambient.b  * illum.src[i].ambient.b + 
						 surf.diffuse.b  * illum.src[i].diffuse.b * foff);

			// specular contribution
			color1.a += att * spot *
						(surf.specular.a * illum.src[i].specular.a * spec);
			color1.r += att * spot *
						(surf.specular.r * illum.src[i].specular.r * spec);
			color1.g += att * spot *
						(surf.specular.g * illum.src[i].specular.g * spec);
			color1.b += att * spot *
						(surf.specular.b * illum.src[i].specular.b * spec);
		}
		else // surface is facing away from this light source
		{
			color0.a += att * spot * surf.ambient.a * illum.src[i].ambient.a;
			color0.r += att * spot * surf.ambient.r * illum.src[i].ambient.r;
			color0.g += att * spot * surf.ambient.g * illum.src[i].ambient.g;
			color0.b += att * spot * surf.ambient.b * illum.src[i].ambient.b;
		}
	}
	
	color0.a = Min(Max(color0.a, 0.f), 1.f);
	color0.r = Min(Max(color0.r, 0.f), 1.f);
	color0.g = Min(Max(color0.g, 0.f), 1.f);
	color0.b = Min(Max(color0.b, 0.f), 1.f);

	color1.a = Min(Max(color1.a, 0.f), 1.f);
	color1.r = Min(Max(color1.r, 0.f), 1.f);
	color1.g = Min(Max(color1.g, 0.f), 1.f);
	color1.b = Min(Max(color1.b, 0.f), 1.f);
}
