#include <math.h>
#include "rendSpan.h"
#include "rendClipRect.h"
#include "rendClipFrust.h"
#include "rendVect.h"
#include "rendPrim.h"
#include "rendHelper.h"

// Trapezoid()			: scan-converts a trapezoid polygon
//		VECTSIZE_T									: must be of R4 or greater
//		VERTEX_T									: one of Vertex* types with the same number of components as VECTSIZE_T
//		SPAN_ROUTINE_T								: one of Span* routines which handles the above VERTEX_T type
//		- y0_p0,		const float (&)[VECTSIZE_T]	: trapezoid 1st vertex,	input
//		- y0_p1,		const float (&)[VECTSIZE_T]	: trapezoid 2nd vertex,	input
//		- y1_p0,		const float (&)[VECTSIZE_T]	: trapezoid 3rd vertex,	input
//		- y1_p1,		const float (&)[VECTSIZE_T]	: trapezoid 4th vertex,	input
// returns
//		nil
// note
//		- vertices must be in ascending-y-ascending-x order
//		- left edge is used for all height calculations
template < int VECTSIZE_T,
		   typename VERTEX_T,
		   void (SPAN_ROUTINE_T)(const VERTEX_T &, const VERTEX_T &) >

void Trapezoid(/*const*/ float (&y0_p0)[VECTSIZE_T], /*const*/ float (&y0_p1)[VECTSIZE_T],
			   /*const*/ float (&y1_p0)[VECTSIZE_T], /*const*/ float (&y1_p1)[VECTSIZE_T])
{
	const float y0 = y0_p0[1];
	const float y1 = y1_p0[1];

	// snap top end to the next greater- or equal integer;
	// snap bottom end to the previous integer
    const float snap_y0 = ceilf(y0);
    const float snap_y1 = ceilf(y1) - 1.f;

    if (snap_y1 < snap_y0)
        return;

    const float err_y0 = snap_y0 - y0;
    const float dx0_dy = (y1_p0[0] - y0_p0[0]) / (y1 - y0);
    const float dx1_dy = (y1_p1[0] - y0_p1[0]) / (y1 - y0);

	const int snapped_h = ToInt(snap_y1 - snap_y0) + 1;
	const int int_y0	= ToInt(snap_y0);

    for (int i = 0; i < snapped_h; i++)
    {
		const float dx0i = (err_y0 + (float) i) * dx0_dy;
		const float dx1i = (err_y0 + (float) i) * dx1_dy;

		float p0i[VECTSIZE_T], p1i[VECTSIZE_T];

		p0i[0] = y0_p0[0] + dx0i;
		p1i[0] = y0_p1[0] + dx1i;

		for (int comp = 2; comp < VECTSIZE_T; comp++)
		{
			p0i[comp] = y0_p0[comp] + (err_y0 + (float) i) * rend_state.trigrad_y[comp - 2] + dx0i * rend_state.trigrad_x[comp - 2];
			p1i[comp] = y0_p1[comp] + (err_y0 + (float) i) * rend_state.trigrad_y[comp - 2] + dx1i * rend_state.trigrad_x[comp - 2];
		}

		rend_state.frame.cursor = rend_state.frame.bits + (int_y0 + i) * rend_state.frame.row_stride;
		rend_state.aux.cursor	= rend_state.aux.bits	+ (int_y0 + i) * rend_state.aux.row_stride;

		SPAN_ROUTINE_T(*(VERTEX_T *)&p0i, *(VERTEX_T *)&p1i);
    }
}

// TriTwiceArea()	: calculates twice the signed area of a 2D triangle, as the length
//					  of the cross-product of two of the sides, taken in the XY plane
//		- p0x,		const float		: X of the triangle's 1st vertex,	input
//		- p0y,		const float		: Y of the triangle's 1st vertex,	input
//		- p1x,		const float		: X of the triangle's 2nd vertex,	input
//		- p1y,		const float		: Y of the triangle's 2nd vertex,	input
//		- p2x,		const float		: X of the triangle's 3rd vertex,	input
//		- p2y,		const float		: Y of the triangle's 3rd vertex,	input
// returns
//		float
//			negative area - counter-clockwise
//			positive area - clockwise
// note
//		- if resulting area's sign will be used as a triangle orienantion determinant,
//		  passed vertices must be in ascending y order
inline float TriTwiceArea(const float p0x, const float p0y,
						  const float p1x, const float p1y,
						  const float p2x, const float p2y)
{
	return (p1x - p0x) * (p2y - p1y) - (p2x - p1x) * (p1y - p0y);
}

// Tri()				: scan-converts a screen-space triangle, performing culling and port clipping
//		VECTSIZE_T									: must be of R4 or greater
//		VERTEX_T									: one of Vertex* types with the same number of components as VECTSIZE_T
//		SPAN_ROUTINE_T								: one of Span* routines which handles the above VERTEX_T type
//		- p0,			const float (&)[VECTSIZE_T]	: triangle 1st vertex,	input
//		- p1,			const float (&)[VECTSIZE_T]	: triangle 2nd vertex,	input
//		- p2,			const float (&)[VECTSIZE_T]	: triangle 3rd vertex,	input
// returns
//		nil
// note
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
template < int VECTSIZE_T,
		   typename VERTEX_T,
		   void (SPAN_ROUTINE_T)(const VERTEX_T &, const VERTEX_T &) >

void Tri(/*const*/ float (&p0)[VECTSIZE_T], /*const*/ float (&p1)[VECTSIZE_T], /*const*/ float (&p2)[VECTSIZE_T])
{
	// find the ascending-Y order of the passed triangle vertices
	const Vertex2 *v_sorted[3];
	v_sorted[0] = (const Vertex2 *) &p0;
	v_sorted[1] = (const Vertex2 *) &p1;
	v_sorted[2] = (const Vertex2 *) &p2;

	uint32 swapped = 0;

	if ((*v_sorted[0]).y > (*v_sorted[1]).y)
	{
		Swap< const Vertex2 * >(v_sorted[0], v_sorted[1]);
		swapped ^= IEEE754_SIGN_MASK;
	}
	if ((*v_sorted[1]).y > (*v_sorted[2]).y)
	{
		Swap< const Vertex2 * >(v_sorted[1], v_sorted[2]);
		swapped ^= IEEE754_SIGN_MASK;

		// vertex1 & vertex2 swapped - re-check vertex0 & vertex1
		if ((*v_sorted[0]).y > (*v_sorted[1]).y)
		{
			Swap< const Vertex2 * >(v_sorted[0], v_sorted[1]);
			swapped ^= IEEE754_SIGN_MASK;
		}
	}

	// calculate the signed area of the triangle
	float triArea = TriTwiceArea((*v_sorted[0]).x, (*v_sorted[0]).y,
								 (*v_sorted[1]).x, (*v_sorted[1]).y,
								 (*v_sorted[2]).x, (*v_sorted[2]).y);

	(*(uint32*)&triArea) ^= swapped;

	// is area close to zero?
	const float eps = (float)1e-15;

	if ((*(uint32*)&triArea & ~IEEE754_SIGN_MASK) < *(const uint32*)&eps)
		return;

	// if culling is enabled reject the triangle if its area's sign is the same as the culling sign
	if (rend_state.culling)
		if (!((*(uint32*)&triArea ^ rend_state.culling) & IEEE754_SIGN_MASK))
			return;

	// clip this triangle by the viewport rectangle
	static const uint32 BUFFSIZE = 40;

	/*const*/ float (*poly[BUFFSIZE])[VECTSIZE_T];
	float buff[BUFFSIZE][VECTSIZE_T];

	poly[0] = &p0;
	poly[1] = &p1;
	poly[2] = &p2;

	uint32 start = 0;
	uint32 size  = 3;
	uint32 nbuff = 0;

	ClipByRect< VECTSIZE_T >(viewport.screen, poly, start, size, buff, nbuff);

	if (size < 3)
		return;

	// triangle has passed all formal rejection tests; reset any per-triangle drawing states
	span_above.x0 = 0;
	span_above.x1 = 0;

	// calculate the gradients of the vertex parameters (triangle gradients)
	float ooTriArea = 1.f / triArea;

	for (int32 g = 0; g < VECTSIZE_T - 2; g++) // iterate the vertex parameters which we need the gradients of
	{
		rend_state.trigrad_x[g] = ooTriArea * TriTwiceArea(p0[2 + g], (*(const Vertex2 *)&p0).y,
														   p1[2 + g], (*(const Vertex2 *)&p1).y,
														   p2[2 + g], (*(const Vertex2 *)&p2).y);

		rend_state.trigrad_y[g] = ooTriArea * TriTwiceArea((*(const Vertex2 *)&p0).x, p0[2 + g],
														   (*(const Vertex2 *)&p1).x, p1[2 + g],
														   (*(const Vertex2 *)&p2).x, p2[2 + g]);
	}

	// find the top-most left and right, and the bottom-most left and right vertices
	int32 vert_ul = start;
	int32 vert_ur = start;
	int32 vert_ll = start;
	int32 vert_lr = start;

	for (uint32 i = start + 1; i < start + size; i++)
	{
		// check for top-most
		if ((*(const Vertex2 *)poly[i]).y < (*(const Vertex2 *)poly[vert_ul]).y) // is this vertex above the top-most?
		{
			vert_ul = i;
			vert_ur = i;
		}
		else
		if ((*(const Vertex2 *)poly[i]).y == (*(const Vertex2 *)poly[vert_ul]).y) // is it at the same y as top-most?
		{
			if ((*(const Vertex2 *)poly[i]).x < (*(const Vertex2 *)poly[vert_ul]).x)
				vert_ul = i;

			if ((*(const Vertex2 *)poly[i]).x > (*(const Vertex2 *)poly[vert_ur]).x)
				vert_ur = i;
		}

		// check for bottom-most
		if ((*(const Vertex2 *)poly[i]).y > (*(const Vertex2 *)poly[vert_ll]).y) // is this vertex below the bottom-most?
		{
			vert_ll = i;
			vert_lr = i;
		}
		else
		if ((*(const Vertex2 *)poly[i]).y == (*(const Vertex2 *)poly[vert_ll]).y) // is it at the same y as bottom-most?
		{
			if ((*(const Vertex2 *)poly[i]).x < (*(const Vertex2 *)poly[vert_ll]).x)
				vert_ll = i;

			if ((*(const Vertex2 *)poly[i]).x > (*(const Vertex2 *)poly[vert_lr]).x)
				vert_lr = i;
		}
	}

	// divide the polygon vertically into non-overlapping trapezoids
	int32 inc_l;

	if (*(uint32 *)&triArea & IEEE754_SIGN_MASK)	// is poly counter-clockwise?
		inc_l = 1;
	else											// poly is clockwise
		inc_l = -1;

	int32 next_l = start + (size + (vert_ul - start) + inc_l) % size;
	int32 next_r = start + (size + (vert_ur - start) - inc_l) % size;

	/*const*/ float (*top_l)[VECTSIZE_T] = poly[vert_ul];
	/*const*/ float (*top_r)[VECTSIZE_T] = poly[vert_ur];

	float (*tmp)[VECTSIZE_T] = buff + nbuff;

	while (vert_ul != vert_ll && vert_ur != vert_lr)
	{
		if ((*(const Vertex2 *)poly[next_l]).y == (*(const Vertex2 *)poly[next_r]).y) // do both segments end simultaneously?
		{
			Trapezoid< VECTSIZE_T, VERTEX_T, SPAN_ROUTINE_T >(*top_l, *top_r,
															  *poly[next_l], *poly[next_r]);
			// symetrical advancement
			vert_ul = next_l;
			vert_ur = next_r;

			next_l = start + (size + (next_l - start) + inc_l) % size;
			next_r = start + (size + (next_r - start) - inc_l) % size;

			top_l = poly[vert_ul];
			top_r = poly[vert_ur];
		}
		else
		if ((*(const Vertex2 *)poly[next_l]).y < (*(const Vertex2 *)poly[next_r]).y) // left segment ends first
		{
			IntersectRow< VECTSIZE_T >(poly[vert_ur], poly[next_r], (*(const Vertex2 *)poly[next_l]).y, tmp);

			Trapezoid< VECTSIZE_T, VERTEX_T, SPAN_ROUTINE_T >(*top_l, *top_r,
															  *poly[next_l], *tmp);
			// advance left
			vert_ul = next_l;

			next_l = start + (size + (next_l - start) + inc_l) % size;

			top_l = poly[vert_ul];
			top_r = tmp++;
		}
		else // right segment ends first
		{
			IntersectRow< VECTSIZE_T >(poly[vert_ul], poly[next_l], (*(const Vertex2 *)poly[next_r]).y, tmp);

			Trapezoid< VECTSIZE_T, VERTEX_T, SPAN_ROUTINE_T >(*top_l, *top_r,
															  *tmp, *poly[next_r]);
			// advance right
			vert_ur = next_r;

			next_r = start + (size + (next_r - start) - inc_l) % size;

			top_l = tmp++;
			top_r = poly[vert_ur];
		}
	}
}

// Triangle()			: draws a triangle, taking it through a model-projection transform,
//						  frustum-clipping it, culling it and port-clipping it
//		VECTSIZE_T									: must be of R4 or greater
//		VERTEX_T									: one of Vertex* types with the same number of components as VECTSIZE_T
//		SPAN_ROUTINE_T								: one of Span* routines which handles the above VERTEX_T type
//		- mat,			const float (&)[4][4]		: model-projection matrix,	input
//		- p0,			const float (&)[VECTSIZE_T]	: triangle 1st vertex,		input
//		- p1,			const float (&)[VECTSIZE_T]	: triangle 2nd vertex,		input
//		- p2,			const float (&)[VECTSIZE_T]	: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
template < int VECTSIZE_T,
		   typename VERTEX_T,
		   void (SPAN_ROUTINE_T)(const VERTEX_T &, const VERTEX_T &) >

void Triangle(const float (&mat)[4][4],
			  /*const*/ float (&p0)[VECTSIZE_T], /*const*/ float (&p1)[VECTSIZE_T], /*const*/ float (&p2)[VECTSIZE_T])
{
	static const uint32 BUFFSIZE = 40;

	float p_clip[BUFFSIZE][VECTSIZE_T];

	Transform4< VECTSIZE_T >(mat, p0, p_clip[0]);
	Transform4< VECTSIZE_T >(mat, p1, p_clip[1]);
	Transform4< VECTSIZE_T >(mat, p2, p_clip[2]);

	const float (*poly[BUFFSIZE])[VECTSIZE_T];

	poly[0] = p_clip + 0;
	poly[1] = p_clip + 1;
	poly[2] = p_clip + 2;

	uint32 start = 0;
	uint32 size = 3;
	uint32 nclip = 3;

	ClipByFrustum< VECTSIZE_T >(poly, start, size, p_clip, nclip);

	if (size < 3)
		return;

	float p_scr[BUFFSIZE][VECTSIZE_T];

	for (uint32 i = 0; i < size; i++)
		MapToScreen< VECTSIZE_T >(*poly[start + i], p_scr[i]);

	// rasterize the resulting polygon as a triangle fan
	for (uint32 i = 0; i < size - 2; i++)
		Tri< VECTSIZE_T, VERTEX_T, SPAN_ROUTINE_T >(p_scr[0], p_scr[i + 1], p_scr[i + 2]);
}

// PlotTriangle4()	: transforms and draws a triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- mat,	const float (&)[4][4]	: model-projection matrix,	input
//		- p0,	const float (&)[4]		: triangle 1st vertex,		input
//		- p1,	const float (&)[4]		: triangle 2nd vertex,		input
//		- p2,	const float (&)[4]		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle4(const float (&mat)[4][4], /*const*/ float (&p0)[4], /*const*/ float (&p1)[4], /*const*/ float (&p2)[4])
{
	if (RENDER_STATE_OP_NOP == rend_state.color_op)
	{
		Triangle< 4, float [4], Span4_nop >(mat, p0, p1, p2);
		return;
	}

	switch (rend_state.vert_format)
	{
	case RENDER_STATE_VERT_BASIC:

		switch((rend_state.color_op << 16) | (rend_state.aux.pix_format << 8) | rend_state.frame.pix_format)
		{
		// msb(0): rend_state.color_op:		  RENDER_STATE_OP_COPY
		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_ARGB1555
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 4, Vertex4, Span4_cpy_ARGB1555_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 4, Vertex4, Span4_cpy_ARGB1555_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 4, Vertex4, Span4_cpy_ARGB1555_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 4, Vertex4, Span4_cpy_ARGB1555_RGB888 >(mat, p0, p1, p2);
			break;

		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_ARGB8888
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 4, Vertex4, Span4_cpy_ARGB8888_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 4, Vertex4, Span4_cpy_ARGB8888_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 4, Vertex4, Span4_cpy_ARGB8888_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 4, Vertex4, Span4_cpy_ARGB8888_RGB888 >(mat, p0, p1, p2);
			break;

		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_RGB565
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 4, Vertex4, Span4_cpy_RGB565_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 4, Vertex4, Span4_cpy_RGB565_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 4, Vertex4, Span4_cpy_RGB565_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 4, Vertex4, Span4_cpy_RGB565_RGB888 >(mat, p0, p1, p2);
			break;

		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_RGB888
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 4, Vertex4, Span4_cpy_RGB888_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 4, Vertex4, Span4_cpy_RGB888_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 4, Vertex4, Span4_cpy_RGB888_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 4, Vertex4, Span4_cpy_RGB888_RGB888 >(mat, p0, p1, p2);
			break;
		}
		break;

		// note: option(s) unavailable for this vertex format:
		//
		// rend_state.color_op:	RENDER_STATE_OP_MODULATE,
		//						RENDER_STATE_OP_DOT3,
		//						RENDER_STATE_OP_ADD
	}
}

// PlotTriangle4()	: draws a screen-space triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- p0,	const float (&)[4]	: triangle 1st vertex,		input
//		- p1,	const float (&)[4]	: triangle 2nd vertex,		input
//		- p2,	const float (&)[4]	: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- vectors' 3rd element (Z) must be in the range [0, 1], and 4th element (RHW) must be non-zero
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle4(/*const*/ float (&p0)[4], /*const*/ float (&p1)[4], /*const*/ float (&p2)[4])
{
	if (RENDER_STATE_OP_NOP == rend_state.color_op)
	{
		Tri< 4, float [4], Span4_nop >(p0, p1, p2);
		return;
	}

	switch (rend_state.vert_format)
	{
	case RENDER_STATE_VERT_BASIC:

		switch((rend_state.color_op << 16) | (rend_state.aux.pix_format << 8) | rend_state.frame.pix_format)
		{
		// msb(0): rend_state.color_op:		  RENDER_STATE_OP_COPY
		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_ARGB1555
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 4, Vertex4, Span4_cpy_ARGB1555_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 4, Vertex4, Span4_cpy_ARGB1555_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 4, Vertex4, Span4_cpy_ARGB1555_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 4, Vertex4, Span4_cpy_ARGB1555_RGB888 >(p0, p1, p2);
			break;

		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_ARGB8888
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 4, Vertex4, Span4_cpy_ARGB8888_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 4, Vertex4, Span4_cpy_ARGB8888_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 4, Vertex4, Span4_cpy_ARGB8888_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 4, Vertex4, Span4_cpy_ARGB8888_RGB888 >(p0, p1, p2);
			break;

		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_RGB565
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 4, Vertex4, Span4_cpy_RGB565_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 4, Vertex4, Span4_cpy_RGB565_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 4, Vertex4, Span4_cpy_RGB565_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 4, Vertex4, Span4_cpy_RGB565_RGB888 >(p0, p1, p2);
			break;

		// msb(1): rend_state.aux.pix_format: RENDER_STATE_PIX_RGB888
		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 4, Vertex4, Span4_cpy_RGB888_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 4, Vertex4, Span4_cpy_RGB888_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 4, Vertex4, Span4_cpy_RGB888_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 4, Vertex4, Span4_cpy_RGB888_RGB888 >(p0, p1, p2);
			break;
		}
		break;

		// note: option(s) unavailable for this vertex format:
		//
		// rend_state.color_op:	RENDER_STATE_OP_MODULATE,
		//						RENDER_STATE_OP_DOT3,
		//						RENDER_STATE_OP_ADD
	}
}

// PlotTriangle6()	: transforms and draws a triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- mat,	const float (&)[4][4]	: model-projection matrix,	input
//		- p0,	const float (&)[6]		: triangle 1st vertex,		input
//		- p1,	const float (&)[6]		: triangle 2nd vertex,		input
//		- p2,	const float (&)[6]		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle6(const float (&mat)[4][4], /*const*/ float (&p0)[6], /*const*/ float (&p1)[6], /*const*/ float (&p2)[6])
{
	if (RENDER_STATE_OP_NOP == rend_state.color_op)
	{
		Triangle< 6, float [6], Span6_nop >(mat, p0, p1, p2);
		return;
	}

	switch (rend_state.vert_format)
	{
	case RENDER_STATE_VERT_UV:

		switch(rend_state.tex.pix_format)
		{
		case RENDER_STATE_PIX_ARGB8888:

			switch((rend_state.color_op << 16) | (rend_state.frame.pix_format << 8) | rend_state.sampling)
			{
			// msb(0): rend_state.color_op:			RENDER_STATE_OP_COPY
			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_ARGB1555
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB1555_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB1555_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB1555_nearest_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB1555_nearest_linear >(mat, p0, p1, p2);
				break;

			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_ARGB8888
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB8888_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB8888_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB8888_nearest_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_ARGB8888_nearest_linear >(mat, p0, p1, p2);
				break;

			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_RGB565
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB565_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB565_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB565_nearest_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB565_nearest_linear >(mat, p0, p1, p2);
				break;

			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_RGB888
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB888_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB888_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB888_nearest_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Triangle< 6, Vertex4UV, Span4UV_cpy_RGB888_nearest_linear >(mat, p0, p1, p2);
				break;
			}
			break;

			// note: option(s) unavailable for this texture format:
			//
			// rend_state.color_op:	RENDER_STATE_OP_MODULATE,
			//						RENDER_STATE_OP_DOT3,
			//						RENDER_STATE_OP_ADD
			//
			// rend_state.sampling:	RENDER_STATE_SAMP_LINEAR_NEAREST,
			//						RENDER_STATE_SAMP_LINEAR_LINEAR

		case RENDER_STATE_PIX_DA16: // rend_state.tex.pix_format

			switch((rend_state.color_op << 16) | (rend_state.sampling << 8) | rend_state.frame.pix_format)
			{
			// msb(0): rend_state.color_op: RENDER_STATE_OP_COPY
			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_CONST_NEAREST
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_ARGB8888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA16_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_RGB888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA12_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_RGB565:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA8_nearest >(mat, p0, p1, p2);
				break;

			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_CONST_LINEAR
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_ARGB8888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA16_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_RGB888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA12_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_RGB565:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA8_linear >(mat, p0, p1, p2);
				break;

			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_NEAREST_NEAREST
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_ARGB8888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA16_nearest_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_RGB888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA12_nearest_nearest >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_RGB565:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA8_nearest_nearest >(mat, p0, p1, p2);
				break;

			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_NEAREST_LINEAR
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_ARGB8888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA16_nearest_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_RGB888:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA12_nearest_linear >(mat, p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_RGB565:

				Triangle< 6, Vertex4UV, Span4UV_cpy_DA8_nearest_linear >(mat, p0, p1, p2);
				break;
			}
			break;

			// note: option(s) unavailable for this texture format:
			//
			// rend_state.color_op:	RENDER_STATE_OP_MODULATE,
			//						RENDER_STATE_OP_DOT3
			//						RENDER_STATE_OP_ADD
			//
			// rend_state.sampling:	RENDER_STATE_SAMP_LINEAR_NEAREST,
			//						RENDER_STATE_SAMP_LINEAR_LINEAR
		}
		break;
	}
}

// PlotTriangle6()	: draws a screen-space triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- p0,	const float (&)[6]	: triangle 1st vertex,		input
//		- p1,	const float (&)[6]	: triangle 2nd vertex,		input
//		- p2,	const float (&)[6]	: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- vectors' 3rd element (Z) must be in the range [0, 1], and 4th element (RHW) must be non-zero
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle6(/*const*/ float (&p0)[6], /*const*/ float (&p1)[6], /*const*/ float (&p2)[6])
{
	if (RENDER_STATE_OP_NOP == rend_state.color_op)
	{
		Tri< 6, float [6], Span6_nop >(p0, p1, p2);
		return;
	}

	switch (rend_state.vert_format)
	{
	case RENDER_STATE_VERT_UV:

		switch(rend_state.tex.pix_format)
		{
		case RENDER_STATE_PIX_ARGB8888:

			switch((rend_state.color_op << 16) | (rend_state.frame.pix_format << 8) | rend_state.sampling)
			{
			// msb(0): rend_state.color_op:			RENDER_STATE_OP_COPY
			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_ARGB1555
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB1555_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB1555_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB1555_nearest_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB1555 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB1555_nearest_linear >(p0, p1, p2);
				break;

			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_ARGB8888
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB8888_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB8888_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB8888_nearest_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_ARGB8888 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_ARGB8888_nearest_linear >(p0, p1, p2);
				break;

			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_RGB565
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB565_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB565_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB565_nearest_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB565 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB565_nearest_linear >(p0, p1, p2);
				break;

			// msb(1): rend_state.frame.pix_format: RENDER_STATE_PIX_RGB888
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_CONST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB888_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_CONST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB888_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_NEAREST_NEAREST:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB888_nearest_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_PIX_RGB888 << 8) | RENDER_STATE_SAMP_NEAREST_LINEAR:

				Tri< 6, Vertex4UV, Span4UV_cpy_RGB888_nearest_linear >(p0, p1, p2);
				break;
			}
			break;

			// note: option(s) unavailable for this texture format:
			//
			// rend_state.color_op:	RENDER_STATE_OP_MODULATE,
			//						RENDER_STATE_OP_DOT3,
			//						RENDER_STATE_OP_ADD
			//
			// rend_state.sampling:	RENDER_STATE_SAMP_LINEAR_NEAREST,
			//						RENDER_STATE_SAMP_LINEAR_LINEAR

		case RENDER_STATE_PIX_DA16: // rend_state.tex.pix_format

			switch((rend_state.color_op << 16) | (rend_state.sampling << 8) | rend_state.frame.pix_format)
			{
			// msb(0): rend_state.color_op: RENDER_STATE_OP_COPY
			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_CONST_NEAREST
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_ARGB8888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA16_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_RGB888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA12_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_NEAREST << 8) | RENDER_STATE_PIX_RGB565:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA8_nearest >(p0, p1, p2);
				break;

			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_CONST_LINEAR
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_ARGB8888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA16_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_RGB888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA12_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_CONST_LINEAR << 8) | RENDER_STATE_PIX_RGB565:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA8_linear >(p0, p1, p2);
				break;

			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_NEAREST_NEAREST
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_ARGB8888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA16_nearest_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_RGB888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA12_nearest_nearest >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_NEAREST << 8) | RENDER_STATE_PIX_RGB565:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA8_nearest_nearest >(p0, p1, p2);
				break;

			// msb(1): rend_state.sampling: RENDER_STATE_SAMP_NEAREST_LINEAR
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_ARGB8888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA16_nearest_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_RGB888:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA12_nearest_linear >(p0, p1, p2);
				break;

			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_ARGB1555:
			case (RENDER_STATE_OP_COPY << 16) | (RENDER_STATE_SAMP_NEAREST_LINEAR << 8) | RENDER_STATE_PIX_RGB565:

				Tri< 6, Vertex4UV, Span4UV_cpy_DA8_nearest_linear >(p0, p1, p2);
				break;
			}
			break;

			// note: option(s) unavailable for this texture format:
			//
			// rend_state.color_op:	RENDER_STATE_OP_MODULATE,
			//						RENDER_STATE_OP_DOT3,
			//						RENDER_STATE_OP_ADD
			//
			// rend_state.sampling:	RENDER_STATE_SAMP_LINEAR_NEAREST,
			//						RENDER_STATE_SAMP_LINEAR_LINEAR
		}
		break;
	}
}

// PlotTriangle8()	: transforms and draws a triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- mat,	const float (&)[4][4]	: model-projection matrix,	input
//		- p0,	const float (&)[8]		: triangle 1st vertex,		input
//		- p1,	const float (&)[8]		: triangle 2nd vertex,		input
//		- p2,	const float (&)[8]		: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- primitive gets clipped by the view frustum
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle8(const float (&mat)[4][4], /*const*/ float (&p0)[8], /*const*/ float (&p1)[8], /*const*/ float (&p2)[8])
{
	if (RENDER_STATE_OP_NOP == rend_state.color_op)
	{
		Triangle< 8, float [8], Span8_nop >(mat, p0, p1, p2);
		return;
	}

	switch(rend_state.vert_format)
	{
	case RENDER_STATE_VERT_ARGB:

		switch((rend_state.color_op << 8) | rend_state.frame.pix_format)
		{
		// msb(0): rend_state.color_op: RENDER_STATE_OP_COPY
		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 8, Vertex4ARGB, Span4ARGB_cpy_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_cpy_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 8, Vertex4ARGB, Span4ARGB_cpy_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_cpy_RGB888 >(mat, p0, p1, p2);
			break;

		// msb(0): rend_state.color_op: RENDER_STATE_OP_ADD
		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 8, Vertex4ARGB, Span4ARGB_add_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_add_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 8, Vertex4ARGB, Span4ARGB_add_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_add_RGB888 >(mat, p0, p1, p2);
			break;

		// msb(0): rend_state.color_op: RENDER_STATE_OP_MODULATE
		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 8, Vertex4ARGB, Span4ARGB_mod_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_mod_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 8, Vertex4ARGB, Span4ARGB_mod_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_mod_RGB888 >(mat, p0, p1, p2);
			break;

		// msb(0): rend_state.color_op: RENDER_STATE_OP_DOT3
		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_ARGB1555:

			Triangle< 8, Vertex4ARGB, Span4ARGB_dot_ARGB1555 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_ARGB8888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_dot_ARGB8888 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_RGB565:

			Triangle< 8, Vertex4ARGB, Span4ARGB_dot_RGB565 >(mat, p0, p1, p2);
			break;

		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_RGB888:

			Triangle< 8, Vertex4ARGB, Span4ARGB_dot_RGB888 >(mat, p0, p1, p2);
			break;
		}
		break;
	}
}

// PlotTriangle8()	: draws a screen-space triangle of stated vertex type, using stated sampling,
//					  and stated pixel format and color operaton
//		- p0,	const float (&)[8]	: triangle 1st vertex,		input
//		- p1,	const float (&)[8]	: triangle 2nd vertex,		input
//		- p2,	const float (&)[8]	: triangle 3rd vertex,		input
// returns
//		nil
// note
//		- vectors' 3rd element (Z) must be in the range [0, 1], and 4th element (RHW) must be non-zero
//		- primitive gets culled according to the render state
//		- primitive gets clipped by the viewport rectangle
void PlotTriangle8(/*const*/ float (&p0)[8], /*const*/ float (&p1)[8], /*const*/ float (&p2)[8])
{
	if (RENDER_STATE_OP_NOP == rend_state.color_op)
	{
		Tri< 8, float [8], Span8_nop >(p0, p1, p2);
		return;
	}

	switch(rend_state.vert_format)
	{
	case RENDER_STATE_VERT_ARGB:

		switch((rend_state.color_op << 8) | rend_state.frame.pix_format)
		{
		// msb(0): rend_state.color_op: RENDER_STATE_OP_COPY
		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 8, Vertex4ARGB, Span4ARGB_cpy_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 8, Vertex4ARGB, Span4ARGB_cpy_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 8, Vertex4ARGB, Span4ARGB_cpy_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_COPY << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 8, Vertex4ARGB, Span4ARGB_cpy_RGB888 >(p0, p1, p2);
			break;

		// msb(0): rend_state.color_op: RENDER_STATE_OP_ADD
		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 8, Vertex4ARGB, Span4ARGB_add_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 8, Vertex4ARGB, Span4ARGB_add_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 8, Vertex4ARGB, Span4ARGB_add_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_ADD << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 8, Vertex4ARGB, Span4ARGB_add_RGB888 >(p0, p1, p2);
			break;

		// msb(0): rend_state.color_op: RENDER_STATE_OP_MODULATE
		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 8, Vertex4ARGB, Span4ARGB_mod_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 8, Vertex4ARGB, Span4ARGB_mod_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 8, Vertex4ARGB, Span4ARGB_mod_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_MODULATE << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 8, Vertex4ARGB, Span4ARGB_mod_RGB888 >(p0, p1, p2);
			break;

		// msb(0): rend_state.color_op: RENDER_STATE_OP_DOT3
		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_ARGB1555:

			Tri< 8, Vertex4ARGB, Span4ARGB_dot_ARGB1555 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_ARGB8888:

			Tri< 8, Vertex4ARGB, Span4ARGB_dot_ARGB8888 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_RGB565:

			Tri< 8, Vertex4ARGB, Span4ARGB_dot_RGB565 >(p0, p1, p2);
			break;

		case (RENDER_STATE_OP_DOT3 << 8) | RENDER_STATE_PIX_RGB888:

			Tri< 8, Vertex4ARGB, Span4ARGB_dot_RGB888 >(p0, p1, p2);
			break;
		}
		break;
	}
}
