#ifndef __RENDVECT__
#define __RENDVECT__

#include <new>
#include <math.h>
#include <memory.h>
#include "rendBaseTypes.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
// remark on conventions
//
// vectors are taken as column-vectors and matrices as column-major, thus
// matrix transformations follow operator-leftism, and are written out as
// Tn * Tn-1 * .. * T1 * T0 * v
////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////
// prototypes
////////////////////////////////////////////////////////////////////////////////////////////////////

void Identity(float (&)[4][4]);
void Rotate(float (&)[4][4], const float, float, float, float);
void Translate(float (&)[4][4], const float, const float, const float);
void Scale(float (&)[4][4], const float, const float, const float);

void Frustum(float (&)[4][4], const float, const float, const float, const float, const float, const float);
void Ortho(float (&)[4][4], const float, const float, const float, const float, const float, const float);

bool InvertMat4(const float (&)[4][4], float (&)[4][4]);

////////////////////////////////////////////////////////////////////////////////////////////////////
// class vector
////////////////////////////////////////////////////////////////////////////////////////////////////

template < int VECTSIZE_T >
class vect
{
	float	m[VECTSIZE_T];

public:

	vect() { }

	static vect< VECTSIZE_T > & adopt(float (&arr)[VECTSIZE_T])
		{ return *new (&arr) vect< VECTSIZE_T >; }

	static const vect< VECTSIZE_T > & adopt(const float (&arr)[VECTSIZE_T])
		{ return *new (& const_cast< float (&)[VECTSIZE_T] >(arr)) vect< VECTSIZE_T >; }

	float & operator [] (const int n)
		{ return m[n]; }

	float operator [] (const int n) const
		{ return m[n]; }

	////////////////////////////////////////////////////////////////////////////////////////////////
	float operator * (const vect< VECTSIZE_T > &v) const;

	float sqr() const
		{ return *this * *this; }

	float norm() const
		{ return sqrtf(sqr()); }

	////////////////////////////////////////////////////////////////////////////////////////////////
	vect< VECTSIZE_T > & operator *= (const float c);

	template < int T >
	friend vect< T > operator * (const vect< T > &v, const float c);

	template < int T >
	friend vect< T > operator * (const float c, const vect< T > &v);

	////////////////////////////////////////////////////////////////////////////////////////////////
	vect< VECTSIZE_T > operator + (const vect< VECTSIZE_T > &v) const;

	vect< VECTSIZE_T > operator - (const vect< VECTSIZE_T > &v) const;
	
	vect< VECTSIZE_T > operator - () const;

	vect< VECTSIZE_T > & operator += (const vect< VECTSIZE_T > &v);

	vect< VECTSIZE_T > & operator -= (const vect< VECTSIZE_T > &v);

	////////////////////////////////////////////////////////////////////////////////////////////////
	vect< VECTSIZE_T > operator ~ () const
		{ return *this * (1.f / norm()); }
};

template < int VECTSIZE_T >
inline float vect< VECTSIZE_T >::operator * (const vect< VECTSIZE_T > &v) const
{
	float r = 0.f;

	for (int i = 0; i < VECTSIZE_T; i++)
		r += m[i] * v.m[i];

	return r;
};

template < int VECTSIZE_T >
inline vect< VECTSIZE_T > & vect< VECTSIZE_T >::operator *= (const float c)
{
	for (int i = 0; i < VECTSIZE_T; i++)
		m[i] *= c;
		
	return *this;
}

template < int VECTSIZE_T >
inline vect< VECTSIZE_T > operator * (const vect< VECTSIZE_T > &v, const float c)
{
	vect< VECTSIZE_T > vr;

	for (int i = 0; i < VECTSIZE_T; i++)
		vr.m[i] = v.m[i] * c;

	return vr;
}

template < int VECTSIZE_T >
inline vect< VECTSIZE_T > operator * (const float c, const vect< VECTSIZE_T > &v)
{
	return v * c;
}

template < int VECTSIZE_T >
inline vect< VECTSIZE_T > vect< VECTSIZE_T >::operator + (const vect< VECTSIZE_T > &v) const
{
	vect< VECTSIZE_T > vr;

	for (int i = 0; i < VECTSIZE_T; i++)
		vr.m[i] = m[i] + v.m[i];

	return vr;
}

template < int VECTSIZE_T >
inline vect< VECTSIZE_T > vect< VECTSIZE_T >::operator - (const vect< VECTSIZE_T > &v) const
{
	vect< VECTSIZE_T > vr;

	for (int i = 0; i < VECTSIZE_T; i++)
		vr.m[i] = m[i] - v.m[i];

	return vr;
}
	
template < int VECTSIZE_T >
inline vect< VECTSIZE_T > vect< VECTSIZE_T >::operator - () const
{
	vect< VECTSIZE_T > vr;

	for (int i = 0; i < VECTSIZE_T; i++)
		vr.m[i] = - m[i];

	return vr;
}

template < int VECTSIZE_T >
inline vect< VECTSIZE_T > & vect< VECTSIZE_T >::operator += (const vect< VECTSIZE_T > &v)
{
	for (int i = 0; i < VECTSIZE_T; i++)
		m[i] += v.m[i];

	return *this;
}

template < int VECTSIZE_T >
inline vect< VECTSIZE_T > & vect< VECTSIZE_T >::operator -= (const vect< VECTSIZE_T > &v)
{
	for (int i = 0; i < VECTSIZE_T; i++)
		m[i] -= v.m[i];

	return *this;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// definitions
////////////////////////////////////////////////////////////////////////////////////////////////////

template < int VECTSIZE_T, int STRIDE0_T, int STRIDE1_T >
inline float DotVect(const float *v0, const float *v1)
{
	float sum = 0.f;
	
	for (int i = 0; i < VECTSIZE_T; i++)
		sum += v0[STRIDE0_T * i] * v1[STRIDE1_T * i];

	return sum;
}

inline void SetMat4(float (&mat)[4][4],
					const float a00, const float a01, const float a02, const float a03,
					const float a10, const float a11, const float a12, const float a13,
					const float a20, const float a21, const float a22, const float a23,
					const float a30, const float a31, const float a32, const float a33)
{
	mat[0][0] = a00;
	mat[0][1] = a01;
	mat[0][2] = a02;
	mat[0][3] = a03;

	mat[1][0] = a10;
	mat[1][1] = a11;
	mat[1][2] = a12;
	mat[1][3] = a13;

	mat[2][0] = a20;
	mat[2][1] = a21;
	mat[2][2] = a22;
	mat[2][3] = a23;

	mat[3][0] = a30;
	mat[3][1] = a31;
	mat[3][2] = a32;
	mat[3][3] = a33;
}

inline void CopyMat4(const float (&src)[4][4], float (&dst)[4][4])
{
	memcpy(&dst, &src, sizeof(float[4][4]));
}

inline void TranspMat4(const float (&mat)[4][4], float (&matT)[4][4])
{
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++) 
			matT[j][i] = mat[i][j];
}

inline void MulMat4(const float (&matA)[4][4], float (&matB)[4][4])
{
	float colB[4][4];

	TranspMat4(matB, colB);

	const vect<4> &rowA0 = vect<4>::adopt(matA[0]);
	matB[0][0] = rowA0 * vect<4>::adopt(colB[0]);
	matB[0][1] = rowA0 * vect<4>::adopt(colB[1]);
	matB[0][2] = rowA0 * vect<4>::adopt(colB[2]);
	matB[0][3] = rowA0 * vect<4>::adopt(colB[3]);

	const vect<4> &rowA1 = vect<4>::adopt(matA[1]);
	matB[1][0] = rowA1 * vect<4>::adopt(colB[0]);
	matB[1][1] = rowA1 * vect<4>::adopt(colB[1]);
	matB[1][2] = rowA1 * vect<4>::adopt(colB[2]);
	matB[1][3] = rowA1 * vect<4>::adopt(colB[3]);

	const vect<4> &rowA2 = vect<4>::adopt(matA[2]);
	matB[2][0] = rowA2 * vect<4>::adopt(colB[0]);
	matB[2][1] = rowA2 * vect<4>::adopt(colB[1]);
	matB[2][2] = rowA2 * vect<4>::adopt(colB[2]);
	matB[2][3] = rowA2 * vect<4>::adopt(colB[3]);

	const vect<4> &rowA3 = vect<4>::adopt(matA[3]);
	matB[3][0] = rowA3 * vect<4>::adopt(colB[0]);
	matB[3][1] = rowA3 * vect<4>::adopt(colB[1]);
	matB[3][2] = rowA3 * vect<4>::adopt(colB[2]);
	matB[3][3] = rowA3 * vect<4>::adopt(colB[3]);
}

inline void MulMat4(const float (&matA)[4][4], const float (&matB)[4][4], float (&matC)[4][4])
{
	matC[0][0] = DotVect<4,1,4>(matA[0], matB[0] + 0);
	matC[0][1] = DotVect<4,1,4>(matA[0], matB[0] + 1);
	matC[0][2] = DotVect<4,1,4>(matA[0], matB[0] + 2);
	matC[0][3] = DotVect<4,1,4>(matA[0], matB[0] + 3);

	matC[1][0] = DotVect<4,1,4>(matA[1], matB[0] + 0);
	matC[1][1] = DotVect<4,1,4>(matA[1], matB[0] + 1);
	matC[1][2] = DotVect<4,1,4>(matA[1], matB[0] + 2);
	matC[1][3] = DotVect<4,1,4>(matA[1], matB[0] + 3);

	matC[2][0] = DotVect<4,1,4>(matA[2], matB[0] + 0);
	matC[2][1] = DotVect<4,1,4>(matA[2], matB[0] + 1);
	matC[2][2] = DotVect<4,1,4>(matA[2], matB[0] + 2);
	matC[2][3] = DotVect<4,1,4>(matA[2], matB[0] + 3);

	matC[3][0] = DotVect<4,1,4>(matA[3], matB[0] + 0);
	matC[3][1] = DotVect<4,1,4>(matA[3], matB[0] + 1);
	matC[3][2] = DotVect<4,1,4>(matA[3], matB[0] + 2);
	matC[3][3] = DotVect<4,1,4>(matA[3], matB[0] + 3);
}

template < int VECTSIZE_T >
inline void Transform3(const float (&mat)[4][4], /*const*/ float (&vi)[VECTSIZE_T], float (&vo)[VECTSIZE_T])
{
	const vect<3> &v = vect<3>::adopt(*(const float (*)[3])&vi);
	vo[0] = vect<3>::adopt(*(const float (*)[3])&mat[0]) * v;
	vo[1] = vect<3>::adopt(*(const float (*)[3])&mat[1]) * v;
	vo[2] = vect<3>::adopt(*(const float (*)[3])&mat[2]) * v;

	for (int comp = 3; comp < VECTSIZE_T; comp++)
		vo[comp] = vi[comp];
}

template < int VECTSIZE_T >
inline void Transform4(const float (&mat)[4][4], /*const*/ float (&vi)[VECTSIZE_T], float (&vo)[VECTSIZE_T])
{
	const vect<4> &v = vect<4>::adopt(*(const float (*)[4])&vi);
	vo[0] = vect<4>::adopt(mat[0]) * v;
	vo[1] = vect<4>::adopt(mat[1]) * v;
	vo[2] = vect<4>::adopt(mat[2]) * v;
	vo[3] = vect<4>::adopt(mat[3]) * v;

	for (int comp = 4; comp < VECTSIZE_T; comp++)
		vo[comp] = vi[comp];
}

template < int VECTSIZE_T >
inline void Interpolate(const float (&p0)[VECTSIZE_T],
						const float (&p1)[VECTSIZE_T],
						float dist0,
						float dist1,
						float (&p_interpol)[VECTSIZE_T])
{
	// take the absolute value of the distances
	*(uint32 *)&dist0 &= ~IEEE754_SIGN_MASK;
	*(uint32 *)&dist1 &= ~IEEE754_SIGN_MASK;
	
	const float mass0 = dist1 / (dist0 + dist1);
	const float mass1 = dist0 / (dist0 + dist1);

	for (int comp = 0; comp < VECTSIZE_T; comp++)
		p_interpol[comp] = p0[comp] * mass0 + p1[comp] * mass1;
}

// ClipByPlane()	: clips a polygon by a plane, 'inside' being the halfspace of positive distances
//		VECTSIZE_T									: must be of R4 or greater
//		- plane,	const float (&)[4]				: halfspaces border,						input
//		- poly,		const float (**)[VECTSIZE_T]	: array containing a polygon definition,	input/output
//		- start,	uint32 &						: polygon starting vertex in the array,		input/output
//		- size,		uint32 &						: number of vertices in the polygon,		input/output
//		- buffer,	float (*)[VECTSIZE_T]			: storage buffer for the new vertices,		input/output
//		- nbuff,	uint32 &						: number of vertices in the storage buffer,	input/output
// returns
//		nil
// note
//		- plane vector defined as (n.x, n.y, n.z, -d), where |n| = 1;
//		- the polygon could be either clockwise or counter-clockwise, orientation is preserved;
//		- result may be a point or a line segment for certain marginal cases
template < int VECTSIZE_T >
inline void ClipByPlane(const float (&plane)[4],
						const float (**poly)[VECTSIZE_T],
						uint32 &start,
						uint32 &size,
						float (*buffer)[VECTSIZE_T],
						uint32 &nbuff)
{
	uint32 currVert	= start;
	uint32 prevVert	= start + size - 1;
	uint32 counter	= size;

	start += size;
	size = 0;

	while (counter--)
	{
		// calculate vertices' distances to the clipping plane
		const vect<4> &vplane = vect<4>::adopt(plane);
		const float distCurr = vect<4>::adopt(*(const float (*)[4])poly[currVert]) * vplane;
		const float distPrev = vect<4>::adopt(*(const float (*)[4])poly[prevVert]) * vplane;

		if (distCurr < 0) // is current vertex outside?
		{
			if (distPrev > 0) // is previous vertex strictly inside?
			{
				Interpolate< VECTSIZE_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}
		}
		else // current vertex is inside
		{
			if (distCurr > 0 &&	// is current vertex strictly inside and
				distPrev < 0)	// is previous vertex outside?
			{
				Interpolate< VECTSIZE_T >(*poly[currVert], *poly[prevVert], distCurr, distPrev, buffer[nbuff]);
				poly[start + size++] = &buffer[nbuff++];
			}

			poly[start + size++] = poly[currVert];
		}

		prevVert = currVert;
		currVert++;
	}
}

#endif // __RENDVECT__
